# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

class RuntimeField:
    def __init__(
            self,
            field_type: str,
            dependencies: list[str],
            function_body: str,
            function_default: str = '',
            params: dict = {},
            **kwargs
    ):
        self.field_type = field_type
        self.dependencies = self.__unpack_dependencies(dependencies)
        self.function_body = function_body
        self.function_default = function_default
        self.params = params
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __unpack_dependencies(self, dependencies):
        ret = []
        for dependency in dependencies:
            if isinstance(dependency, RuntimeField):
                ret.extend(dependency.dependencies)
            else:
                ret.append(dependency)
        return ret

    def __if_dependencies_exist(self):
        conditions = [
            f"doc.containsKey('{dep}') && "
            f"doc['{dep}'].size() > 0" for dep in self.dependencies
        ]
        if self.dependencies:
            return (
                f"if ({' && '.join(conditions)})"
            )
        return 'if (1==1)'

    def to_function(self):
        main_body = f"""
            {self.__if_dependencies_exist()} {{
                {self.function_body}
            }}
        """
        if self.function_default:
            main_body += f"""
                else {{
                    {self.function_default}
                }}
            """
        return f'{main_body}'

    def to_dict(self):
        return {
            'type': self.field_type,
            'script': {
                'source': self.to_function(),
                'params': self.params
            }
        }


class RuntimeFields:
    @classmethod
    def date_interval(cls, start_date: str, end_date: str, unit: str = 'days'):
        rf = RuntimeField(
            field_type='long',
            dependencies=[start_date, end_date],
            function_body=f"""
                ZonedDateTime start = doc['{start_date}'].value;
                ZonedDateTime end = doc['{end_date}'].value;
                long differenceInMillis = ChronoUnit.{unit.upper()}.between(start, end);
                emit(differenceInMillis);
            """
        )
        return rf.to_dict()

    @classmethod
    def latest_date(cls, date_fields: list[str], allow_missing: bool = True):
        deps = date_fields if not allow_missing else []
        rf = RuntimeField(
            field_type='date',
            dependencies=deps,
            params={'dates': date_fields},
            function_body="""
                ZonedDateTime latestDate = null;
                for (int i = 0; i < params.dates.size(); i++) {
                    String dep = params.dates.get(i);
                    if (doc.containsKey(dep) && doc[dep].size() > 0) {
                        ZonedDateTime currentDate = doc[dep].value;
                        if (latestDate == null || currentDate.isAfter(latestDate)) {
                            latestDate = currentDate;
                        }
                    }
                }
                if (latestDate != null) {
                    emit(latestDate.toInstant().toEpochMilli());
                }
            """
        )
        return rf.to_dict()

    @classmethod
    def math(cls, first: str, second: str, operation: str = '/',
             return_type: str = 'double'):
        rf = RuntimeField(
            field_type=return_type,
            dependencies=[first, second],
            function_body=f"""
                emit(doc['{first}'].value {operation} doc['{second}'].value)
            """
        )
        return rf.to_dict()

    @classmethod
    def coalesce(cls, fields: list[str], return_type: str = 'keyword'):
        keyword = '.keyword' if return_type == 'keyword' else ''
        rf = RuntimeField(
            field_type=return_type,
            dependencies=[],  # No dependencies as we check in function_body
            function_body='else '.join(
                [
                    f"if (doc.containsKey('{field}{keyword}') "
                    f"&& doc['{field}{keyword}'].size() > 0) {{"
                    f"emit(doc['{field}{keyword}'].value);"
                    f'}}'
                    for field in fields
                ]
            )
        )
        return rf.to_dict()



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os
from typing import Optional

import requests
from requests.adapters import HTTPAdapter, Retry


class HttpClient:
    """
    Core functionality for HTTP clients
    """

    def __init__(
        self,
        token: Optional[str] = None,
        token_header: str = 'token',
        retries: int = 5,
        status_forcelist: Optional[list[int]] = [429, 502, 503, 504],
    ) -> None:
        self.__token = self._token_header(token_header, token)
        self.__retries = retries
        self.__status_forcelist = status_forcelist

    def _token_header(
        self,
        key: str,
        token: Optional[str],
    ) -> Optional[dict[str, str]]:

        return None if token is None else {key: token}

    def _merge_headers(
        self,
        headers: Optional[dict[str, str]] = None
    ) -> dict[str, str]:
        """
        Merges (possibly `None`) headers with the
        `Optional[str]` token. Returns `None` if both are
        undefined
        """

        def __empty_if_none(
            d: Optional[dict[str, str]]
        ) -> dict[str, str]:
            return {} if d is None else d

        if self.__token is None and headers is None:
            return None
        return {
            **__empty_if_none(headers),
            **__empty_if_none(self.__token)
        }

    def get_session(self) -> requests.Session:

        cert_path = os.path.join(
            os.path.dirname(__file__),
            '..',
            '..',
            'certs',
            'cacert.pem'
        )

        session = requests.Session()
        session.verify = cert_path

        return session

    def _get_session_with_retries(self) -> requests.Session:
        """
        Attempts a call to the endpoint 5 times, with a delay of 1 second
        """
        session = self.get_session()

        retry_strategy = Retry(
            total=self.__retries,
            backoff_factor=1,
            status_forcelist=self.__status_forcelist
        )
        session.mount('http://', HTTPAdapter(max_retries=retry_strategy))
        session.mount('https://', HTTPAdapter(max_retries=retry_strategy))

        return session



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Optional

import requests

from .converter import LabwhereApiTransfer


class LabwhereApiClient:
    """
    Takes LabWhere API transfers and connects to a remote
    LabWhere API.
    """

    def __init__(
        self,
        labwhere_url: str,
    ) -> None:
        self.__labwhere_url = labwhere_url

    def get_detail(
        self,
        object_type: str,
        object_id: str
    ) -> Optional[LabwhereApiTransfer]:
        """
        Gets a single Labwhere API transfer for the object of specified
        `object_type` and `object_id`, or returns None if not found.
        """

        url = self.__detail_url(object_type, object_id)
        return self.__fetch_detail(url)

    def __fetch_detail(
        self,
        url: str
    ) -> Optional[LabwhereApiTransfer]:

        r = requests.get(url)
        if r.status_code == 404:
            return None
        r.raise_for_status()
        return r.json()

    def __detail_url(self, object_type: str, object_id: str) -> str:
        return f'{self.__labwhere_url}/{object_type}s/{object_id}'



# SPDX-FileCopyrightText: 2026 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from datetime import datetime, timezone
from typing import Any

from .action import Action
from ..core import DataSource


class SetStatusAction(Action):
    """
    Sets the status of one or more objects by inserting a new row into
    the corresponding status history table.
    """

    def run(
        self,
        datasource: DataSource,
        ids: list[str],
        object_type: str,
        params: dict[str, Any] | None = None
    ) -> tuple[dict[str, bool], int]:

        if not params or 'status' not in params:
            return {'error': 'Missing required param: "status"'}, 400

        status_type_id = params['status']
        status_table = f'{object_type}_status'
        status_type_table = f'{object_type}_status_type'

        try:
            status_type = datasource.get_one(status_type_table, status_type_id)

            new_status_objects = self.__build_status_objects(
                datasource=datasource,
                ids=ids,
                object_type=object_type,
                status_table=status_table,
                status_type=status_type,
            )

            with datasource.get_session() as session:
                for obj in session.insert(status_table, new_status_objects):
                    parent = obj.to_one_relationships.get(object_type)
                    setattr(parent, 'status', obj)
                    session.upsert(object_type, [parent])

            return {'success': True}, 200
        except Exception as e:  # noqa: BLE001
            return {'error': str(e)}, 500

    def __build_status_objects(
        self,
        datasource: DataSource,
        ids: list[str],
        object_type: str,
        status_table: str,
        status_type: Any,
    ) -> Any:

        for id_ in ids:
            parent = datasource.get_one(object_type, id_)
            yield datasource.data_object_factory(
                type_=status_table,
                attributes={
                    'status_time': datetime.now(tz=timezone.utc),
                },
                to_one={
                    object_type: parent,
                    'status_type': status_type,
                },
            )



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from datetime import datetime
from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class ElasticSequencingRequestToStsSequencingRequestConverter(
        DataObjectToDataObjectOrUpdateConverter):

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        submission_date = datetime.fromtimestamp(0)
        if 'benchling_completion_date' in data_object.attributes and \
                data_object.benchling_completion_date is not None:
            submission_date = data_object.benchling_completion_date

        ret = self._data_object_factory(
            'sequencing_request',
            data_object.id,
            attributes={
                'fluidx_id': data_object.benchling_fluidx_id,
                'platform': data_object.benchling_sequencing_platform.upper(),
                'submission_date': submission_date
            })
        return iter([ret])



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from functools import cache
from typing import Callable, Iterable, Optional

from .client import CopoApiClient
from .converter import (
    CopoApiConverter
)
from ..core import DataObject, DataSource, DataSourceError, DataSourceFilter
from ..core.operator import (
    DetailGetter,
    ListGetter,
    Relational
)
from ..core.relationship import RelationshipConfig

if typing.TYPE_CHECKING:
    from ..core.session import OperableSession


ClientFactory = Callable[[], CopoApiClient]
CopoConverterFactory = Callable[[], CopoApiConverter]


class CopoDataSource(
    DataSource,

    # the supported operators
    DetailGetter,
    ListGetter,
    Relational
):
    """
    A `DataSource` that connects to a remote COPO API.

    Developers should likely use `create_copo_datasource`
    instead of this directly.
    """

    def __init__(
        self,
        client_factory: ClientFactory,
        copo_converter_factory: CopoConverterFactory
    ) -> None:

        self.__client_factory = client_factory
        self.__lc_factory = copo_converter_factory
        super().__init__({})

    @property
    @cache
    def attribute_types(self) -> dict[str, dict[str, str]]:
        return {
            'manifest': {
            },
            'sample': {
                'associated_tol_project': 'str',
                'biosampleAccession': 'str',
                'copo_id': 'str',
                'copo_profile_title': 'str',
                'manifest_id': 'str',
                'manifest_version': '',
                'public_name': 'str',
                'sampleDerivedFrom': 'str',
                'sampleSameAs': 'str',
                'sampleSymbiontOf': 'str',
                'sraAccession': 'str',
                'status': 'str',
                'submissionAccession': 'str',
                'time_created': 'datetime',
                'time_updated': 'datetime',
                'tol_project': 'str'
            }
        }

    @property
    @cache
    def relationship_config(self) -> dict[str, RelationshipConfig]:
        rc_sample = RelationshipConfig()
        rc_sample.to_one = {
            'manifest': 'manifest'
        }
        rc_manifest = RelationshipConfig()
        rc_manifest.to_many = {'samples': 'sample'}
        return {
            'sample': rc_sample,
            'manifest': rc_manifest
        }

    @property
    @cache
    def supported_types(self) -> list[str]:
        return list(
            self.attribute_types.keys()
        )

    def get_by_id(
        self,
        object_type: str,
        object_ids: Iterable[str],
        **kwargs,
    ) -> Iterable[Optional[DataObject]]:
        if object_type not in self.supported_types:
            raise DataSourceError(f'{object_type} is not supported')

        client = self.__client_factory()
        copo_responses = client.get_detail(object_type, object_ids)
        copo_converter = self.__lc_factory()
        return (
            copo_converter.convert(r)
            if r is not None else None
            for r in copo_responses
        )

    def get_list(
        self,
        object_type: str,
        object_filters: Optional[DataSourceFilter] = None,
        session: Optional[OperableSession] = None
    ) -> Iterable[DataObject]:
        """
        This is a very limited implementation of `get_list`.
        It only supports checking for manifests in a project
        between certain dates
        """
        client = self.__client_factory()
        jc_converter = self.__lc_factory()

        project = object_filters.and_ \
            .get('project', {}) \
            .get('eq', {}) \
            .get('value', None)
        from_ = object_filters.and_ \
            .get('time_created', {}) \
            .get('gte', {}) \
            .get('value', None)
        to = object_filters.and_ \
            .get('time_created', {}) \
            .get('lt', {}) \
            .get('value', None)

        manifests_json = client.get_manifests(
            project=project,
            from_=from_,
            to=to,
        )
        manifests, _ = jc_converter.convert_list(manifests_json)
        return manifests

    def get_to_one_relation(
        self,
        source: DataObject,
        relationship_name: str
    ) -> Optional[DataObject]:

        if source.type == 'sample' and relationship_name == 'manifest':
            if source.manifest_id is not None:
                return self.get_one(
                    'manifest',
                    source.manifest_id
                )

    def get_to_many_relations(
        self,
        source: DataObject,
        relationship_name: str
    ) -> Iterable[DataObject]:
        if source.type == 'manifest' and relationship_name == 'samples':
            client = self.__client_factory()
            copo_responses = client.get_samples_in_manifest(source.id)
            copo_converter = self.__lc_factory()
            return (
                copo_converter.convert(r)
                if r is not None else None
                for r in copo_responses
            )



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import datetime
import importlib
import json
import os
import re
import shlex
import subprocess

import click

from dotenv import load_dotenv

from ..core import (
    DataSourceFilter
)


@click.group()
@click.option(
    '--env-file',
    default='.env.dev',
    type=click.Path(),
    help='Use a custom environment file.',
    show_default=True,
)
def cli(env_file):
    pass


@cli.command()
@click.option(
    '--type',
    'type_',
    default='python',
    type=click.Choice(['python', 'license', 'ui', 'ui-fix']),
    help='Type of lint.',
    show_default=True,
)
def lint(type_):
    """
    Run linting.
    """

    click.echo('Running lint...')
    if type_ == 'license':
        command = 'docker run --rm --volume $(pwd):/data fsfe/reuse:1.1.2 lint'
        click.secho(command, fg='green')
        run(command)
    if type_ == 'python':
        linter = 'gitlab-registry.internal.sanger.ac.uk/tol/tol-core/lint:1.0.3'
        command = f'docker run --rm --volume $(pwd):/project {linter}'
        click.secho(command, fg='green')
        run(command)
    if type_ == 'ui':
        ui_linter = 'gitlab-registry.internal.sanger.ac.uk/tol/tol-core/ui-lint:1.0.9'
        command = f'docker run --rm --volume $(pwd):/src {ui_linter}'
        click.secho(command, fg='green')
        run(command)
    if type_ == 'ui-fix':
        ui_linter = 'gitlab-registry.internal.sanger.ac.uk/tol/tol-core/ui-lint:1.0.9 '
        command_1 = f'docker run --rm --volume $(pwd):/src {ui_linter}'
        command_2 = 'npx eslint \
            -c /project/.eslintrc \
            --ext .js,.jsx,.ts,.tsx \
            --fix \
            --ignore-pattern "**/public" \
            --ignore-pattern "**/*.license" \
            --ignore-pattern "**/Dockerfile" \
            --ignore-pattern "**/*.dev" \
            --ignore-pattern "**/*.scss" \
            /src/*-ui/**/*'
        click.secho((command_1 + command_2), fg='green')
        run(command_1 + command_2)


@cli.command()
@click.option(
    '--type',
    'type_',
    default='sast',
    type=click.Choice(['sast', 'dependencies']),
    help='Type of scan to run.',
    show_default=True,
)
def scan(type_):
    """
    Perform a Snyk security scan.
    """
    token = os.getenv('SNYK_TOKEN')
    if token is None:
        click.secho('SNYK_TOKEN environment variable must be set', fg='red')
        return
    # service = get_app()
    click.echo('Running scan...')
    snyc_common = 'docker run --env SNYK_TOKEN --rm --volume $(pwd):/app'
    if type_ == 'sast':
        command = f'{snyc_common} snyk/snyk:python snyk code test'
    elif type_ == 'dependencies':
        command = f'{snyc_common} snyk/snyk:python snyk test'
    click.secho(command, fg='green')
    run(command)


@cli.command()
@click.option('--ui/--no-ui', default=True, help='Build and run the UI container.')
@click.option('--db/--no-db', default=True, help='Build and run the DB container.')
@click.option('--api/--no-api', default=True, help='Build and run the API container.')
@click.pass_context
def up(ctx, ui, db, api):
    """
    Start ToL services.
    """
    env_file = ctx.parent.params['env_file']
    service = get_app()
    click.echo(f'Starting {service}...')
    containers = []
    if ui:
        containers.append(f'{service}-ui')
    if db:
        containers.append(f'{service}-db')
    if api:
        containers.append(f'{service}-api')
    command = f'docker compose --env-file {env_file} up --build --detach ' + ' '.join(containers)
    click.secho(command, fg='green')
    run(command)
    if api:
        click.secho('API: ' + ' '.join(get_container_urls(f'{service}-api')), fg='yellow')
    if ui:
        click.secho(
            'UI: ' + ' '.join(get_container_urls(f'{service}-ui', protocol='https')), fg='yellow'
        )


@cli.command()
@click.option('--ui/--no-ui', default=True, help='Log the UI container.')
@click.option('--db/--no-db', default=True, help='Log the DB container.')
@click.option('--api/--no-api', default=True, help='Log the API container.')
@click.pass_context
def log(ctx, ui, db, api):
    """
    Log ToL services.
    """
    env_file = ctx.parent.params['env_file']
    service = get_app()
    containers = []
    if ui:
        containers.append(f'{service}-ui')
    if db:
        containers.append(f'{service}-db')
    if api:
        containers.append(f'{service}-api')
    command = (
        f'docker compose --env-file {env_file} logs --tail=0 --follow '
        + ' '.join(containers)
    )
    click.secho(command, fg='green')
    run(command)


@cli.command()
@click.pass_context
def down(ctx):
    """
    Stop ToL services.
    """
    env_file = ctx.parent.params['env_file']
    service = get_app()
    click.echo(f'Stopping {service}...')
    command = f'docker compose  --env-file {env_file} down'
    click.secho(command, fg='green')
    run(command)


@cli.command()
def prune():
    """
    Prune all non-running Docker containers and volumes.
    """
    click.echo('Pruning all Docker containers, volumes, etc...')
    command = 'docker system prune -af'
    click.secho(command, fg='green')
    run(command)
    command = 'docker volume prune -f'
    click.secho(command, fg='green')
    run(command)


@cli.command()
@click.pass_context
def restore(ctx):
    """
    Restore a database from backup.
    """
    env_file = ctx.parent.params['env_file']
    service = get_app()
    click.echo('Restoring database...')
    command = (
        f'docker compose --env-file {env_file} run --rm {service}-dbutils'
        ' restore'
    )
    click.secho(command, fg='green')
    run(command)


@cli.group
@click.pass_context
def alembic(ctx):
    """
    Run Alembic.
    """


@alembic.command()
@click.pass_context
def upgrade(ctx):
    """
    Run migrations on the database.
    """
    env_file = ctx.parent.parent.params['env_file']
    service = get_app()
    click.echo('Running alembic upgrade...')
    command = (
        f'docker compose build {service}-api && docker compose --env-file {env_file}'
        f' run --rm {service}-alembic alembic upgrade head'
    )
    click.secho(command, fg='green')
    run(command)


@alembic.command()
@click.option('--message', required=True, help='migration message')
@click.pass_context
def migration(ctx, message):
    """
    Create a new database migration.
    """
    env_file = ctx.parent.parent.params['env_file']
    service = get_app()
    click.echo('Creating alembic migration...')
    command = (
        f'docker compose build {service}-api && docker compose --env-file {env_file}'
        f' run --rm {service}-alembic alembic revision -m "{message}"'
    )
    click.secho(command, fg='green')
    run(command)


@alembic.command()
@click.pass_context
def merge(ctx):
    """
    Merge heads.
    """
    env_file = ctx.parent.parent.params['env_file']
    service = get_app()
    click.echo('Merging heads...')
    command = (
        f'docker compose build {service}-api && docker compose --env-file {env_file}'
        f' run --rm {service}-alembic alembic merge heads -m "merge heads"'
    )
    click.secho(command, fg='green')
    run(command)


@cli.command(
    context_settings={
        # For passing unknown options to pytest
        'ignore_unknown_options': True,
    }
)
@click.option(
    '--type',
    'type_',
    default='unit',
    type=click.Choice(
        [
            'unit',
            'system',
            'integration',
            'ui',
            'playwright',
        ]
    ),
    help='Type of test to run.',
    show_default=True,
)
@click.argument(
    'pytest_options',
    nargs=-1,
    type=click.UNPROCESSED,
)
@click.pass_context
def test(ctx, type_, pytest_options):
    """
    Run tests. Any unrecognised options on the command line are accumulated in
    `PYTEST_OPTIONS` and passed to `pytest` for the "unit", "system"
    and "integration" test types.

    For example, to run just the system test `my_test`:

       tol test --type system -k my_test
    """
    env_file = ctx.parent.params['env_file']
    service = get_app()

    click.echo('Running tests...')
    if type_ in {'unit', 'system', 'integration'}:
        verbosity = '-v' if type_ == 'unit' else '-vv'
        # shlex will correctly re-quote any quoted arguments from the command
        # line:
        pytest_cmd = shlex.join(['pytest', verbosity, type_, *pytest_options])
        pytest_sh = (
            f'sh -c "if [ -d {type_} ]; then {pytest_cmd};'
            f' else echo \'No {type_} tests found\'; fi"'
        )
        if type_ == 'unit':
            docker_compose_entry = f'{service}-python-unit-test'
            command = (
                f'docker compose build {docker_compose_entry}'
                f' && docker compose --env-file {env_file}'
                f' run --rm {docker_compose_entry} {pytest_sh}'
            )
        elif type_ == 'system':
            docker_compose_entry = f'{service}-python-system-test'
            db_entry = f'{service}-python-db'
            command = (
                f'docker compose build {docker_compose_entry}'
                f' && docker compose --env-file {env_file} up -d {db_entry}'
                f' && docker compose --env-file {env_file}'
                f' run --rm --build {docker_compose_entry} {pytest_sh}'
            )
        elif type_ == 'integration':
            docker_compose_entry = f'{service}-python-integration-test'
            command = (
                f'docker compose build {docker_compose_entry}'
                f' && docker compose --env-file {env_file}'
                f' run --rm --build {docker_compose_entry} {pytest_sh}'
            )
    elif type_ == 'ui':
        docker_compose_entry = f'{service}-ui-test'
        command = (
            f'docker compose build {docker_compose_entry}'
            f' && docker compose --env-file {env_file}'
            f' run --rm {docker_compose_entry} npm run test'
        )
    elif type_ == 'playwright':
        docker_compose_entry = f'{service}-playwright-test'
        command = (
            f'docker compose build {docker_compose_entry}'
            f' && docker compose --env-file {env_file} '
            f' run --rm {docker_compose_entry} npx playwright test'
        )
    click.secho(command, fg='green')
    run(command)


@cli.command()
@click.argument('filename', type=click.Path(exists=True))
@click.pass_context
def flow(ctx, filename):
    """
    Run a flow.
    """
    entry = f'{get_app()}-flow'
    env_file = ctx.parent.params['env_file']
    click.echo('Running flow...')
    command = (
        f'docker compose --env-file {env_file} run --rm --build '
        f'{entry} python3 /opt/prefect/flows/{filename}'
    )
    click.secho(command, fg='green')
    run(command)


@cli.command()
@click.option(
    '--source',
    default='portal',
    type=click.Choice(
        [
            'portal',
            'goat',
            'grit',
            'sts',
            'tolid',
            'tolqc',
            'workflows',
        ]
    ),
    help='Source DataSource',
    show_default=True,
)
@click.option(
    '--operation',
    default='list',
    type=click.Choice(['list']),
    help='Operation to run.',
    show_default=True,
)
@click.option(
    '--type',
    'type_',
    required=True,
    help='Object type.',
)
@click.option(
    '--filter',
    'filter_',
    default=None,
    help='Filter to apply.',
)
@click.option(
    '--fields',
    default='',
    help='Fields to output.',
)
@click.option(
    '--converter',
    default=None,
    help='Converter function,',
)
@click.option(
    '--output',
    default='ndjson',
    type=click.Choice(
        ['ndjson', 'tsv'],
    ),
    help='Format of the output.',
    show_default=True,
)
@click.pass_context
def data(ctx, source, operation, type_, filter_, fields, converter, output):
    """
    Fetch data from a DataSource.
    """
    env_file = ctx.parent.params['env_file']
    if os.path.exists(env_file):
        load_dotenv(ctx.parent.params['env_file'])
    module = importlib.import_module(f'tol.sources.{source}')
    class_ = getattr(module, source)
    ds = class_()
    f = DataSourceFilter()
    if filter_ is not None:
        try:
            provided_filter = json.loads(filter_)
            f.and_ = provided_filter['and_'] if 'and_' in provided_filter else {}
        except json.JSONDecodeError:
            pass
    if operation == 'list':
        objs = ds.get_list(type_, object_filters=f)
    if converter is not None:
        module = importlib.import_module('tol.flows.converters')
        class_ = getattr(module, converter)
        objs = class_(ds.data_object_factory).convert_iterable(objs)
    if output == 'tsv':
        output_tsv(objs, fields.split(',') if fields else [])
    if output == 'ndjson':
        output_json(objs, fields.split(',') if fields else [])


def output_tsv(objs, fields):
    for i, obj in enumerate(objs):
        if i == 0:
            if len(fields) == 0:
                fields = obj.attributes.keys()
            click.echo('\t'.join(fields))

        click.echo('\t'.join(
            str(obj.get_field_by_name(field))
            for field in fields
        ))


def output_json(objs, fields):
    """
    This is a very simple implementation. It doesn't yet handle relationships.
    """
    def datetime_handler(x):
        if isinstance(x, datetime.datetime):
            return x.isoformat()
        raise TypeError('Unknown type')

    for obj in objs:
        click.echo(json.dumps(
            {
                'id': obj.id,
                **obj.attributes
            },
            default=datetime_handler)
        )


def get_app():
    return os.path.basename(os.getcwd())


def run(command):
    try:
        subprocess.run(
            ['bash', '-i', '-c', command]
        ).check_returncode()
    except subprocess.CalledProcessError as e:
        exit(e.returncode)


def run_capture(command):
    s = subprocess.run(['bash', '-i', '-c', command], check=True, capture_output=True)
    return s.stdout.decode('utf-8')


def get_container_ids(name_prefix):
    ids = []
    output = run_capture('docker container ls')
    for line in output.split('\n'):
        if re.search(name_prefix, line):
            ids.append(line.split()[0])
    return ids


def get_container_urls(name_prefix, protocol='http'):
    urls = []
    container_ids = get_container_ids(name_prefix)
    for container_id in container_ids:
        if container_id != '':
            mapping = run_capture(f'docker container port {container_id}')
            if mapping != '':
                urls.append(f'{protocol}://' + mapping.split()[2])
    return urls



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import Any, Dict, List, Optional, Type

from sqlalchemy import distinct, func
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Load, MappedColumn, Query, Session, joinedload, load_only, raiseload
from sqlalchemy.orm.attributes import flag_modified

from .filter import DatabaseFilter
from .model import Model
from .session import SessionFactory
from .sort import DatabaseSorter
from ..core import DataSourceError, ReqFieldsTree


SubPath = tuple[str | None, ReqFieldsTree]


class Database(ABC):
    """Encapsulates basic operations on a Database"""

    @abstractmethod
    def get_by_id(
        self,
        tablename: str,
        instance_id: Any,
        in_session: Session,
        requested_tree: ReqFieldsTree | None = None,
    ) -> Optional[Model]:
        """
        Gets a single instance by its instance-ID, or None if not found.

        Note that this "instance-ID" may not always be named "id" on the
        `Model` class.
        """

    @abstractmethod
    def get_page(
        self,
        tablename: str,
        in_session: Session,
        filters: Optional[DatabaseFilter] = None,
        sort_by: Optional[DatabaseSorter] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> Iterable[Model]:
        """
        Returns an Iterable of `Model` instances according
        to the given filters, offset, and limit.
        """

    @abstractmethod
    def count(
        self,
        tablename: str,
        in_session: Session,
        filters: Optional[DatabaseFilter] = None
    ) -> int:
        """
        Counts the total number of `Model` instances of the given
        tablename matching the given filters.
        """

    @abstractmethod
    def delete(
        self,
        tablename: str,
        instance_id: Any,
        in_session: Session,
        user_id: Optional[str] = None
    ) -> None:
        """
        Deletes the `Model` instance of specified tablename and
        instance-ID.
        """

    @abstractmethod
    def upsert(
        self,
        instance: Model,
        in_session: Session,
        user_id: Optional[str] = None
    ) -> Model:
        """Performs an "upsert" on the given `Model` instance."""

    @abstractmethod
    def insert(
        self,
        instance: Model,
        in_session: Session,
        user_id: Optional[str] = None
    ) -> Model:
        """
        "Inserts" the given `Model` instance.

        Must not already exist.
        """

    @abstractmethod
    def get_to_one_relation(
        self,
        tablename: str,
        instance_id: str,
        relationship_name: str,
        in_session: Session
    ) -> Optional[Model]:
        """
        For the instance of given tablename and ID, gets the to-one
        instance under the given relationship.
        """

    @abstractmethod
    def get_to_many_relations(
        self,
        tablename: str,
        instance_id: str,
        relationship_name: str,
        in_session: Session
    ) -> Iterable[Model]:
        """
        For the instance of given tablename and ID, gets the to-many
        instances under the given relationship.
        """

    @abstractmethod
    def get_group_stats(
        self,
        tablename: str,
        group_by: list[str],
        stats_fields: list[str],
        stats: list[str],
        in_session: Session,
        filters: DatabaseFilter | None = None,
    ) -> list[dict[str, dict[str, Any]]]:
        """
        Provides the specified `stats`:

        - grouped by the cartesian product of unique tuples
          under `group_by`
        - on the fields specified in `stats_fields`
        """

    @property
    @abstractmethod
    def attribute_types(self) -> dict[str, dict[str, type]]:
        """
        The mapping of attribute name to type for each model under
        this `Database` instance.
        """

    @property
    @abstractmethod
    def attribute_types_including_id(self) -> dict[str, dict[str, type]]:
        """
        The mapping of attribute name to type for each model under
        this `Database` instance, including the ID attribute.
        """

    @property
    @abstractmethod
    def session_factory(self) -> SessionFactory:
        """
        Returns a `Callable` that returns a `Session` instance.
        """


class DefaultDatabase(Database):
    """A reasonable-default implementation of the Database ABC."""

    def __init__(
        self,
        session_factory: SessionFactory,
        models: List[Type[Model]]
    ) -> None:

        self.__session_factory = session_factory
        self.__tablename_model_dict = self.__get_tablename_model_dict(models)
        self.__attribute_types_including_id = self.__get_attribute_types_including_id()
        self.__attribute_types = self.__get_attribute_types()

    @property
    def session_factory(self) -> SessionFactory:
        return self.__session_factory

    def get_by_id(
        self,
        tablename: str,
        instance_id: Any,
        in_session: Session,
        requested_tree: ReqFieldsTree | None = None,
    ) -> Optional[Model]:
        result = self.__get_instance_by_id(
            tablename,
            instance_id,
            in_session,
            requested_tree,
        )
        return result

    def get_page(
        self,
        tablename: str,
        in_session: Session,
        filters: Optional[DatabaseFilter] = None,
        sort_by: Optional[DatabaseSorter] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> Iterable[Model]:

        _, query = self.__get_model_query(
            tablename,
            in_session,
            requested_tree=requested_tree,
            filters=filters,
        )
        if filters is not None:
            if sort_by is not None:
                filters.add_field(sort_by.term)
            query = filters.filter(query, tablename, self.__tablename_model_dict)
        if sort_by is not None:
            query = sort_by.sort(query, tablename, self.__tablename_model_dict, filters)
        if limit is not None and offset is not None:
            query = query.limit(limit).offset(offset)
        results = query.all()
        return results

    def count(
        self,
        tablename: str,
        in_session: Session,
        filters: Optional[DatabaseFilter] = None
    ) -> int:

        _, query = self.__get_model_query(tablename, in_session, filters=filters)
        if filters is not None:
            query = filters.filter(query, tablename, self.__tablename_model_dict)
        count = query.count()
        return count

    def delete(
        self,
        tablename: str,
        instance_id: Any,
        in_session: Session,
        user_id: Optional[str] = None
    ) -> None:

        instance = self.__get_instance_by_id(
            tablename,
            instance_id,
            in_session
        )
        in_session.delete(instance)
        return self.__commit_session(
            in_session,
            instance,
            'deletion',
            user_id=user_id,
            is_delete=True
        )

    def upsert(
        self,
        instance: Model,
        in_session: Session,
        user_id: Optional[str] = None,
        merge_collections: bool | None = None,
    ) -> Model:

        instance = self.__upsert_to_session(
            instance,
            in_session,
            merge_collections,
        )
        in_session.flush()
        return self.__commit_session(
            in_session,
            instance,
            'upserting',
            user_id=user_id
        )

    def insert(
        self,
        instance: Model,
        in_session: Session,
        user_id: Optional[str] = None,
    ) -> Model:

        in_session.add(instance)

        return self.__commit_session(
            in_session,
            instance,
            'inserting',
            user_id=user_id
        )

    def get_to_one_relation(
        self,
        tablename: str,
        instance_id: str,
        relationship_name: str,
        in_session: Session,
    ) -> Optional[Model]:

        instance = self.__get_instance_by_id(
            tablename,
            instance_id,
            in_session
        )
        result = instance.instance_to_one_relations[relationship_name]
        return result

    def get_to_many_relations(
        self,
        tablename: str,
        instance_id: str,
        relationship_name: str,
        in_session: Session
    ) -> Iterable[Model]:

        instance = self.__get_instance_by_id(
            tablename,
            instance_id,
            in_session
        )
        result = instance.instance_to_many_relations[relationship_name]
        return result

    def get_group_stats(
        self,
        tablename: str,
        group_by: list[str],
        stats_fields: list[str],
        stats: list[str],
        in_session: Session,
        filters: DatabaseFilter | None = None,
    ) -> list[dict[str, dict[str, Any]]]:

        model, query = self.__get_model_query(
            tablename,
            in_session,
            filters=filters,
        )

        query = self.__filter_query(query, tablename, filters)
        query, s_columns = self.__apply_stats(
            query,
            filters,
            model,
            stats_fields,
            stats,
        )
        query, g_columns = self.__apply_group_by(
            query,
            model,
            group_by,
            filters,
        )

        return self.__get_stats_from_query(
            query,
            group_by,
            stats_fields,
            stats,
            g_columns,
            s_columns,
        )

    @property
    def attribute_types(self) -> dict[str, dict[str, type]]:
        return self.__attribute_types

    @property
    def attribute_types_including_id(self) -> dict[str, dict[str, type]]:
        return self.__attribute_types_including_id

    def __get_stats_from_query(
        self,
        query: Query,
        group_by: list[str],
        stats_fields: list[str],
        stats: list[str],
        group_by_columns: list[MappedColumn],
        stat_columns: list[MappedColumn],
    ) -> list[dict[str, dict[str, Any]]]:

        query = query.with_entities(
            *group_by_columns,
            *stat_columns,
        )
        results = query.all()

        return self.__parse_results(
            results,
            group_by,
            stats_fields,
            stats,
        )

    def __parse_results(
        self,
        results: list[list[Any]],
        group_by: list[str],
        stats_fields: list[str],
        stats: list[str],
    ) -> list[dict[str, dict[str, Any]]]:

        return [
            self.__parse_row(
                r,
                group_by,
                stats_fields,
                stats,
            )
            for r in results
        ]

    def __parse_row(
        self,
        result: list[Any],
        group_by: list[str],
        stats_fields: list[str],
        stats: list[str],
    ) -> dict[str, dict[str, Any]]:

        group_keys = self.__parse_group_keys(
            result,
            group_by,
        )
        stats_dict = self.__parse_stats_dict(
            result,
            group_by,
            stats_fields,
            stats,
        )

        return {
            'key': group_keys,
            'stats': stats_dict,
        }

    def __parse_stats_dict(
        self,
        result: list[Any],
        group_by: list[str],
        stats_fields: list[str],
        stats: list[str],
    ) -> dict[str, Any]:

        stats_dict = {}

        num_stats = len(stats)
        stats_values = result[len(group_by):]

        for i, field in enumerate(stats_fields):
            values = stats_values[i * num_stats:(i + 1) * num_stats]
            values_dict = dict(zip(stats, values))
            stats_dict[field] = values_dict

        stats_dict['count'] = stats_values[-1]

        return stats_dict

    def __parse_group_keys(
        self,
        result: list[Any],
        group_by: list[str],
    ) -> dict[str, Any]:

        group_values = result[:len(group_by)]

        return dict(zip(group_by, group_values))

    def __apply_stats(
        self,
        query: Query,
        filters: DatabaseFilter,
        model: type[Model],
        stats_fields: list[str],
        stats: list[str],
    ) -> tuple[Query, list[MappedColumn]]:

        stat_columns = []

        for field in stats_fields:
            column = self.__get_column(
                filters,
                field,
            )
            stat_columns.extend(
                self.__apply_stats_for_field(
                    column,
                    stats,
                )
            )

        stat_columns.append(func.count())

        return query, stat_columns

    def __apply_stats_for_field(
        self,
        column: MappedColumn,
        stats: list[str],
    ) -> list[Any]:

        return [
            self.__get_stat_clause(
                column,
                stat,
            )
            for stat in stats
        ]

    def __get_column(
        self,
        filters: DatabaseFilter,
        field: str
    ) -> MappedColumn:

        column = filters.get_column(field)
        if '.' in field:
            filters.add_field(field)

        return column

    def __get_stat_clause(
        self,
        column: MappedColumn,
        stat: str,
    ) -> Any:

        if stat == 'min':
            return func.min(column)
        elif stat == 'max':
            return func.max(column)
        elif stat == 'sum':
            # don't try to sum datetimes
            if column.type.python_type not in [int, float]:
                return None
            return func.sum(column)
        elif stat == 'unique':
            return func.count(distinct(column))

    def __apply_group_by(
        self,
        query: Query,
        model: type[Model],
        group_by: list[str],
        filters: DatabaseFilter,
    ) -> tuple[Query, list[MappedColumn]]:
        """Must be done last in `get_group_stats()`."""

        g_columns = []

        for g in group_by:
            g_column = self.__get_column(
                filters,
                g
            )
            g_columns.append(g_column)

        return query.group_by(*g_columns).order_by(*g_columns), g_columns

    def __filter_query(
        self,
        query: Query,
        tablename: str,
        filters: DatabaseFilter | None,
    ) -> Query:

        if filters:
            query = filters.filter(
                query,
                tablename,
                self.__tablename_model_dict,
            )

        return query

    def __get_attribute_types(self) -> dict[str, dict[str, type]]:
        return {
            t: {
                k: v for k, v in self.__attribute_types_including_id[t].items()
                if k != 'id'  # Exclude ID attribute
            }
            for t in self.__attribute_types_including_id.keys()
        }

    def __get_attribute_types_including_id(self) -> dict[str, dict[str, type]]:
        return {
            t: m.get_attribute_types() | {'id': m.get_id_attribute_type()}
            for t, m in self.__tablename_model_dict.items()
        }

    def __get_model_query(
        self,
        tablename: str,
        in_session: Session,
        requested_tree: ReqFieldsTree | None = None,
        filters: DatabaseFilter | None = None,
    ) -> tuple[Type[Model], Query]:

        model = self.__tablename_model_dict[tablename]
        query = filters.get_query(in_session, model) if filters else in_session.query(model)

        if requested_tree:
            query = self.add_options_to_query(query, tablename, requested_tree)

        return model, query

    def __commit_session(
        self,
        in_session: Session,
        instance: Model,
        operation: str,
        user_id: Optional[str] = None,
        is_delete: bool = False
    ) -> Model | None:

        try:
            if not is_delete:
                self.__before_commit(instance, in_session, user_id)
            in_session.commit()
            if not is_delete:
                in_session.refresh(instance)
                return instance
        except IntegrityError:
            in_session.rollback()
            self.__raise_integrity_error(instance, operation)

    def __before_commit(
        self,
        instance: Model,
        in_session: Session,
        user_id: Optional[str]
    ) -> None:

        instance.before_commit(user_id=user_id)
        in_session.merge(instance)

    def __get_instance_by_id(
        self,
        tablename: str,
        instance_id: str,
        in_session: Session,
        requested_tree: ReqFieldsTree | None = None,
    ) -> Optional[Model]:
        """
        Gets an instance by its tablename and id.
        """

        model, query = self.__get_model_query(
            tablename,
            in_session,
            requested_tree=requested_tree,
        )
        id_column = getattr(model, model.get_id_column_name())
        result = query.filter(id_column == instance_id).one_or_none()
        return result

    def __get_tablename_model_dict(
        self,
        models: List[Type[Model]]
    ) -> Dict[str, Type[Model]]:

        return {
            m.get_table_name(): m for m in models
        }

    def __upsert_to_session(
        self,
        instance: Model,
        in_session: Session,
        merge_collections: bool | None,
    ) -> Model:

        old_instance = self.__get_instance_by_id(
            instance.get_table_name(),
            instance.instance_id,
            in_session
        )
        if old_instance is not None:
            instance = self.__handle_difference(
                old_instance,
                instance,
                merge_collections,
            )
            in_session.flush()
            return instance

        in_session.add(instance)
        return instance

    def __handle_difference(
        self,
        old: Model,
        new: Model,
        merge_collections: bool | None,
    ) -> Model:
        """
        Performs various pre-merge operations, including:

        - merging `list`s and `dict`s like `ElasticDataSource`.
        """

        if merge_collections is not False:
            self.__handle_diff_dict(old, new)
            self.__handle_diff_list(old, new)

        __keys = [
            *new.get_attribute_types(),
            *new.get_all_foreign_key_names(),
        ]

        for k in __keys:
            new_val = getattr(new, k)
            if self.__ignore_new_val(new, k, new_val):
                continue
            setattr(old, k, new_val)

        return old

    def __ignore_new_val(
        self,
        instance: Model,
        k: str,
        v: Any
    ) -> bool:

        """
        Returns `True` if the value is both:

        - equal to `None`
        - not previously directly set on the Model
        """

        return v is None and not self.__explititly_set(
            instance,
            k
        )

    def __explititly_set(
        self,
        instance: Model,
        k: str
    ) -> bool:

        if not hasattr(instance, '_sa_instance_state'):
            return False

        return k in instance._sa_instance_state.dict

    def __handle_diff_list(self, old: Model, new: Model) -> None:
        for k in self.__get_type_column_keys(new, list):
            merge_list = getattr(old, k, [])
            if merge_list is None:
                continue

            new_list = getattr(new, k, [])

            for el in new_list:
                if el not in merge_list:
                    merge_list.append(el)

            setattr(new, k, merge_list)

            flag_modified(old, k)

    def __handle_diff_dict(self, old: Model, new: Model) -> None:
        for k in self.__get_type_column_keys(new, dict):
            merge_dict = getattr(old, k, {})
            if merge_dict is None:
                continue

            new_dict = getattr(new, k, {})

            setattr(
                new,
                k,
                merge_dict | new_dict
            )

            flag_modified(old, k)

    def __get_type_column_keys(
        self,
        instance: Model,
        type_: type[Any]
    ) -> list[str]:

        return [
            k for k in instance.get_attribute_types().keys()
            if self.__is_type(instance, k, type_)
        ]

    def __is_type(
        self,
        instance: Model,
        key: str,
        type_: type[Any]
    ) -> bool:
        value = getattr(instance, key)
        return isinstance(value, type_)

    def __raise_integrity_error(
        self,
        instance: Model,
        operation_name: str
    ) -> None:
        relationship_values = instance.get_to_many_relationship_config().values()
        relationship_names = ', '.join(relationship_values)
        raise DataSourceError(
            title='Database Integrity Error',
            detail=(
                'An integrity error was encountered in the Database during '
                f'{operation_name} of the row with tablename '
                f'"{instance.get_table_name()}" and instance-ID '
                f'"{instance.instance_id}". This is usually due '
                f'to another instance pointing towards this one. '
                f'Hint - check the following tables: "{relationship_names}".'
            )
        )

    def add_options_to_query(
        self,
        query: Query,
        tablename: str,
        requested_tree: ReqFieldsTree,
    ):
        options = self.joinedload_options(requested_tree)
        # `raiseload(*)` acts as a trap, raising an exception if any methods
        # on the returned objects are called which would trigger loading data
        # from the database via another SELECT.
        return query.options(options, raiseload('*')) if options else query

    def joinedload_options(self, tree: ReqFieldsTree) -> list[Load]:
        """
        Returns a list of SQLAlchemy `Load` objects based on the supplied
        `ReqFieldsTree` which specify which related tables to join into and
        which attribute columns to select.  This list of `Load` objects can
        then be added to a SQLAlchemy `Query` via a call to `options()`.
        """
        sub: SubPath = None, tree
        return list(self.__joinedload_iter(sub))

    def __joinedload_iter(self, sub: SubPath, *path: list[SubPath]):
        path = [*path, sub]
        tree = sub[1]
        if tree.is_leaf:
            if options := self.__joinedload_options_from_path(path):
                yield options
        else:
            for sub in tree.sub_trees():
                yield from self.__joinedload_iter(sub, *path)

    def __joinedload_options_from_path(self, path: list[SubPath]):
        load = None
        prev_model = None
        for rel_name, tree in path:
            model = self.__tablename_model_dict[tree.object_type]
            if prev_model:
                # Add a joinedload for this model if `tree` isn't the root
                relation = getattr(prev_model, rel_name)
                load = load.joinedload(relation) if load else joinedload(relation)
            if names := tree.attribute_names:
                # SQLAlchemy will always add the primary key to the query, but
                # if the `.id` column isn't the primary key then
                # serialization to JSON will fail with the `raiseload` trap.
                # Adding it unconditionally here is harmless - it won't
                # appear twice in the SELECT statement.
                names.append(model.get_id_column_name())

                for col_name in model.get_all_foreign_key_names():
                    # Always load any to-one ID columns where the relation
                    # isn't being fetched so that we can create stub objects
                    # for them.
                    if not tree.has_attribute(col_name):
                        names.append(col_name)
                cols = [getattr(model, x) for x in names if x != 'id']
                load = (
                    load.load_only(*cols, raiseload=True)
                    if load
                    else load_only(*cols, raiseload=True)
                )
            prev_model = model

        return load



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod
from typing import Dict, Optional, Type

from sqlalchemy.orm import MappedColumn, Query

from .filter import DatabaseFilter
from .model import Model


class DatabaseSorter(ABC):
    """Runs order_by against a query"""

    @abstractmethod
    def sort(
        self,
        query: Query,
        tablename: str,
        model_dict: Dict[str, Type[Model]],
        filters: DatabaseFilter,
    ) -> Query:
        """Sorts a query using the given models"""


class DefaultDatabaseSorter(DatabaseSorter):

    def __init__(self, sort_term: Optional[str]) -> None:
        if sort_term is None:
            self.__desc = False
            self.term = 'id'
        elif sort_term.startswith('-'):
            self.__desc = True
            self.term = sort_term[1:]
        else:
            self.__desc = False
            self.term = sort_term

    def sort(
        self,
        query: Query,
        tablename: str,
        model_dict: Dict[str, Type[Model]],
        filters: DatabaseFilter | None = None,
    ) -> Query:

        if filters is not None:
            column = filters.get_column(self.term)
        else:
            column, query = self.__join_and_get_column(
                query,
                model_dict[tablename],
                model_dict,
            )
        return self.__apply_sort(query, column)

    def __join_and_get_column(
        self,
        query: Query,
        base_model: type[Model],
        model_dict: Dict[str, Type[Model]]
    ) -> tuple[MappedColumn, Query]:

        model = base_model

        relations = self.term.split('.')[:-1]
        for relation in relations:
            column = getattr(model, relation)
            query = query.join(column)
            to_one = model.get_to_one_relationship_config()
            model = model_dict[
                to_one[relation]
            ]

        column = self.__get_column(
            model,
            self.term.split('.')[-1]
        )

        return column, query

    def __get_column(self, model: Type[Model], term: str) -> MappedColumn:
        if term == 'id':
            id_key = model.get_id_column_name()
            return model.get_column(id_key)
        else:
            return model.get_column(term)

    def __apply_sort(self, query: Query, column: MappedColumn) -> Query:
        if self.__desc:
            return query.order_by(column.desc())
        else:
            return query.order_by(column)



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from datetime import datetime
from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class ElasticSampleToStsSampleConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        eln_updated_at = data_object.sts_eln_updated_at
        if not eln_updated_at \
                and data_object.benchling_eln_tissue_id is not None:
            eln_updated_at = datetime.now()
        yield self._data_object_factory(
            'sample',
            data_object.id,
            attributes={
                'public_name':
                    data_object.tolid_tolid.id
                    if data_object.tolid_tolid else None,
                'eln_id': data_object.benchling_eln_tissue_id,
                'ep_exported': True if data_object.benchling_eln_tissue_id else False,
                'eln_updated_at': eln_updated_at,
            }
        )



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class TolqcSampleToElasticSequencingRequestConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        target_attributes = {**data_object.attributes}
        target_to_one = {}

        if data_object.specimen is not None:
            target_to_one['tolid'] = self._data_object_factory(
                'tolid',
                data_object.specimen.id
            )

            target_to_one['specimen'] = self._data_object_factory(
                'specimen',
                data_object.specimen.supplied_name
            )
            if data_object.specimen.accession is not None:
                target_attributes['biospecimen_id'] = data_object.specimen.accession.id
            if data_object.specimen.species is not None:
                target_to_one['species'] = self._data_object_factory(
                    'species',
                    data_object.specimen.species.taxon_id
                )
        ret = self._data_object_factory(
            'sequencing_request',
            data_object.id,
            attributes=target_attributes,
            to_one=target_to_one
        )

        yield ret



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Any, Protocol

if typing.TYPE_CHECKING:
    from .operator import Relational
    from .datasource import DataSource


class _AnyKeyProtocol(Protocol):
    """
    Type hints that every key is valid for an attribute.

    Useful for `DataObject`.
    """

    def __getattr__(self, name: str) -> Any | None:
        ...


DataDict = dict[str, Any]


class DataObject(_AnyKeyProtocol, ABC):
    """
    The ABC for the unit of data on which a DataSource instance
    operates - representing the lingua franca of DataSource instances,
    and declaring all abstract properties that are needed.
    """

    @property
    @abstractmethod
    def type(self) -> str:  # noqa
        """
        The type of this object (e.g. species/specimens/samples).
        """

    @property
    @abstractmethod
    def id(self) -> str | None:  # noqa
        """
        A unique ID by which to identify this object within
        its type.
        """

    @property
    @abstractmethod
    def provenance(self) -> str | None:
        """
        The original source of this object (used in Elastic where the same data can be stored
        in multiple places. In this case, source priority is used to know which to return)
        """

    @property
    @abstractmethod
    def attributes(self) -> dict[str, Any]:
        """
        A dictionary of key:attribute pairs, where an attribute
        is any entry on the object that is none of an ID, type,
        or relationship.
        """

    @property
    @abstractmethod
    def to_one_relationships(self) -> dict[str, DataObject | None]:
        """
        A dictionary of relationships, where this object refers to
        precisely one other.
        """

    @property
    @abstractmethod
    def to_many_relationships(self) -> dict[str, Iterable[DataObject]]:
        """
        A dictionary of relationships, where many objects refer to
        precisely this object.
        """

    @property
    @abstractmethod
    def _host(self) -> DataSource | Relational:
        """
        The DataSource instance that manages DataObject instances of this type
        """

    @property
    @abstractmethod
    def _to_one_objects(self) -> dict[str, DataObject | None]:
        """
        The name: attribute mapping for to-one `DataObject`s set on this
        instance.

        Can used to inspect which relations are set on the object without
        triggering auto-fetching of to-one related objects from the `_host`
        `DataSource`.

        Most users should use `to_one_relationships` instead.
        """

    @property
    @abstractmethod
    def _to_many_objects(self) -> dict[str, Iterable[DataObject]]:
        """
        The name: attribute mapping for to-many `DataObject`s set on this
        instance.

        Can used to inspect which relations are set on the object without
        triggering auto-fetching of to-many related objects from the `_host`
        `DataSource`.

        Most users should use `to_many_relationships` instead.
        """

    def get_field_by_name(self, field_name: str) -> Any:
        """
        Get a field by name, or return `None` if the field does not exist.
        """
        # Split by dots to allow for nested fields
        field_names = field_name.split('.')
        current_obj = self
        for name in field_names:
            current_obj = getattr(current_obj, name)
            if current_obj is None:
                return None
        return current_obj


@dataclass(frozen=True)
class ErrorObject:
    """
    Returned by write `Operator` methods, in place of a valid `DataObject`,
    if there was an error with "writing" an individual input `DataObject` in
    the `Iterable`.
    """

    details: dict[str, Any]
    """Additional detail on this error"""
    object_type: str
    """The `type` of the object that this write concerns"""

    error_id: str | None = None
    """An optional ID for this error"""
    object_id: str | None = None
    """The `id` of the object that this write concerns, if provided"""
    object_: DataObject | None = None
    """The object that this write concerns, if provided"""
    http_code: int | None = None
    """An optional HTTP Status Code for this error"""


WriteObject = DataObject | ErrorObject



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import itertools
import typing
from abc import ABC
from typing import Any, Iterable, Optional

import more_itertools

from ._writer import _Writer

if typing.TYPE_CHECKING:
    from ..data_object import DataObject, ErrorObject
    from ..session import OperableSession


class Upserter(_Writer, ABC):
    """
    Upserts DataObject instances.
    """
    def upsert(
        self,
        object_type: str,
        objects: Iterable[DataObject],
        session: Optional[OperableSession] = None,
        **kwargs: Any
    ) -> Iterable[DataObject | ErrorObject] | None:
        """
        Takes a type and an `Iterable` of `DataObject` instances, on
        each of which to perform an "upsert", i.e.:

        - insert    - if the `DataObject` is new to the `DataSource`
        - update    - if the `DataObject` is not new

        We previde a default implementation that calls `upsert_batch`
        """

        upserted = ()
        for batch in more_itertools.chunked(objects, self.write_batch_size):
            upserted = itertools.chain(upserted, self.upsert_batch(
                object_type,
                batch,
                session=session,
                **kwargs
            ))
        return upserted

    def upsert_arbitary_type(
        self,
        objects: Iterable[DataObject],
        session: Optional[OperableSession] = None,
        **kwargs: Any,
    ) -> Iterable[DataObject | ErrorObject]:
        """
        Calls `upsert()` internally, using contiguous slices of
        `DataObject` instances with the same `type`.

        Unlike with other `upsert*` methods, you always need to
        exhuast the iterable returned by this method (e.g. by a `for`
        loop or using `list()`).
        """

        split_iter = more_itertools.split_when(
            objects,
            lambda x, y: x.type != y.type,
        )

        upserted_iter = (
            self.upsert(
                split[0].type,
                split,
                session=session,
            )
            for split in split_iter
        )

        return itertools.chain.from_iterable(
            upserted
            for upserted in upserted_iter
            if upserted is not None
        )

    def upsert_batch(
        self,
        object_type: str,
        objects: Iterable[DataObject],
        session: Optional[OperableSession] = None,
        **kwargs: Any,
    ) -> Iterable[DataObject | ErrorObject] | None:
        raise NotImplementedError()



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


from functools import cache
from typing import Dict, Iterable, Optional

from ..core import (
    DataObject,
    core_data_object
)
from ..core.operator import Relational
from ..core.relationship import RelationshipConfig
from ..json import S3JsonDataSource


class GapDataSource(
    S3JsonDataSource,
    Relational
):
    def __init__(
        self,
        config: Dict,
        secure: bool = True
    ) -> None:
        super().__init__(
            config=config,
            secure=secure
        )

    @property
    @cache
    def relationship_config(self) -> dict[str, RelationshipConfig]:
        rc_assembly = RelationshipConfig()
        rc_assembly.to_many = {
            'pipelines': 'pipeline',
            'assembly_details': 'assembly_detail'
        }
        return {
            'assembly': rc_assembly
        }

    def get_to_one_relation(
        self,
        source: DataObject,
        relationship_name: str
    ) -> Optional[DataObject]:
        return None

    def get_to_many_relations(
        self,
        source: DataObject,
        relationship_name: str,
    ) -> Iterable[DataObject]:
        if relationship_name == 'pipelines':
            temp_config = {
                'uri': f's3://gap/{source.id}/data/analysis.json',
                'type': 'pipeline',
                'id_attribute': 'pipeline',
                'mappings': {
                    'pipeline': {
                        'heading': 'pipeline',
                        'type': 'str'
                    },
                    'analysis': {
                        'heading': 'analysis',
                        'type': 'str'
                    },
                    'results': {
                        'heading': 'results',
                        'type': 'str'
                    },
                    'description': {
                        'heading': 'description',
                        'type': 'str'
                    },
                    's3': {
                        'heading': 's3',
                        'type': 'str'
                    },
                    'lustre_path_analysis': {
                        'heading': 'lustre_path_analysis',
                        'type': 'str'
                    },
                },
                's3_host': 'cog.sanger.ac.uk',
                's3_access_key': None,
                's3_secret_key': None
            }
            temp_ds = S3JsonDataSource(config=temp_config, secure=True)
            cdo = core_data_object(temp_ds) # noqa
            return temp_ds.get_list('pipeline')

        elif relationship_name == 'assembly_details':
            temp_config = {
                'uri': f's3://gap/{source.id}/data/assembly.json',
                'type': 'assembly_detail',
                'id_attribute': 'name',
                'mappings': {
                    'name': {
                        'heading': 'name',
                        'type': 'str'
                    },
                    'info': {
                        'heading': 'info',
                        'type': 'str'
                    },
                },
                's3_host': 'cog.sanger.ac.uk',
                's3_access_key': None,
                's3_secret_key': None
            }
            temp_ds = S3JsonDataSource(config=temp_config, secure=True)
            cdo = core_data_object(temp_ds) # noqa
            return temp_ds.get_list('assembly_detail')



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from collections.abc import Iterable, Iterator
from functools import reduce
from itertools import chain
from typing import (
    Any,
    Generic,
    TypeVar
)

if typing.TYPE_CHECKING:
    from .data_loader import DataLoader


In = TypeVar('In')
"""The input representation type"""


Out = TypeVar('Out')
"""The output representation type"""


class Converter(ABC, Generic[In, Out]):
    """
    A useful `ABC` for converting from one representation to another.
    """

    def convert_iterable(
        self,
        inputs: Iterable[In | None]
    ) -> Iterable[Out | None]:
        """
        Converts an `Iterable` of (possibly `None`) input representations
        to an `Iterable` of (possibly `None`) output representations,
        according to the rules of `convert_optional()`
        """

        return (self.convert_optional(i) for i in inputs)

    def convert_optional(self, input_: In | None) -> Out | None:
        """
        Converts a possibly `None` input representation to either:

        - `None` if the input is `None`
        - `convert(input)` if the input is not `None`
        """

        return self.convert(input_) if input_ is not None else None

    @abstractmethod
    def convert(self, input_: In) -> Out:
        """
        Converts an input representation to an output representation.

        If the input could be `None`, use `convert_optional()` instead.
        """


class AsyncConverter(ABC, Generic[In, Out]):
    """An asynchronous version of `Converter`"""

    @abstractmethod
    async def async_convert(self, input_: In) -> Out:
        """input_ must not be null."""

    async def async_convert_optional(self, input_: In | None) -> Out | None:
        """input_ can be `None`, in which case `None` is returned"""

        return await self.async_convert(input_) if input_ is not None else None

    async def async_convert_iterable(
        self,
        input_: Iterable[In | None]
    ) -> Iterable[Out | None]:
        """
        Converts an `Iterable`, converting or returning `None` element-wise,
        using `AsyncConverter().async_convert_optional()`
        """

        return [
            await self.async_convert_optional(i) for i in input_
        ]


def is_iter(in_: Any) -> bool:
    try:
        iter(in_)
        return True
    except TypeError:
        return False


class ChainedConverter(Converter, Generic[In, Out]):
    """
    Using multiple converters in sequence, converts in a chain.

    e.g. with `Converter` instances, `a` and `b`, this is roughly
    equivalent to:

    ```
    def convert(self, input_: In) -> Out:
        inner = a.convert(input_)
        return b.convert(inner)
    ```

    Please note that this is different to the main `Converter` -
    `None` elements are ignored in the sequence. Only use the
    `convert_iterable` method.
    """

    def __init__(
        self,
        *converters: Converter
    ):
        self.__converters = converters
        self.__data_loader = None

    @property
    def data_loader(self) -> DataLoader | None:
        return self.__data_loader

    @data_loader.setter
    def data_loader(self, dl: DataLoader) -> None:
        self.__data_loader = dl
        for converter in self.__converters:
            converter.data_loader = dl

    def convert_iterable(
        self,
        inputs: Iterable[In | None]
    ) -> Iterator[Out]:

        return chain.from_iterable(
            self.convert_optional(input_)
            for input_ in inputs
        )

    def convert_optional(
        self,
        input_: In | None
    ) -> Iterator[Out]:

        if input_ is None:
            return iter([])
        else:
            return self.convert(input_)

    def convert(self, input_: In) -> Iterator[Out]:
        return reduce(
            self.__convert_with,
            self.__converters,
            iter([input_])
        )

    def __convert_with(
        self,
        previous: Iterator[Any],
        converter: Converter
    ) -> Iterator[Any]:

        for p in previous:
            converted = converter.convert(p)

            if is_iter(converted):
                yield from converted
            else:
                yield converted

# not ideal, but this is a workaround
    def get_return_objects(self) -> Iterable[Out]:
        return []



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import re
from typing import Dict, Optional

from .filter_utils import FilterUtils
from ...api_client.exception import BadQueryArgError


class ListGetParameters:
    """
    Parses the parameters from a query string for a List GET
    endpoint.
    """

    def __init__(self, request_args: Dict[str, str]) -> None:
        self.__request_args = request_args

    @property
    def page_size(self) -> Optional[int]:
        """The optional page size to return"""
        page_size = self.__request_args.get('page_size')
        if page_size is None:
            return None

        return self.__parse_to_positive_int('page_size', page_size)

    @property
    def page(self) -> Optional[int]:
        """
        The optional number of the page of results.
        """
        page_number = self.__request_args.get('page')
        if page_number is None:
            return None

        return self.__parse_to_positive_int('page', page_number)

    @property
    def filter(self) -> Optional[str]:  # noqa A003
        """
        The optional filter JSON string.
        """
        filter_string = self.__request_args.get('filter')
        if filter_string is None:
            return None

        return FilterUtils.parse_to_datasource_filter('filter', filter_string)

    @property
    def sort_by(self) -> Optional[str]:
        """
        The optional column to sort by.
        """
        sort_by = self.__request_args.get('sort_by')
        if sort_by is None:
            return None

        return self.__parse_to_sort_by_string('sort_by', sort_by)

    @property
    def merge_collections(self) -> bool | None:
        """
        Whether or not to merge dict and list properties when updating
        records.
        """
        merge_cllctns = self.__request_args.get('merge_collections')
        if merge_cllctns is None:
            return None
        return self.__parse_to_boolean('merge_collections', merge_cllctns)

    @property
    def requested_fields(self) -> list[str] | None:
        """
        The list of requested fields.
        """

        fields = self.__request_args.get('requested_fields')
        if fields is None:
            return None

        return self.__parse_to_list_str(fields)

    def __parse_to_list_str(self, __value: str) -> list[str]:
        return list(
            __value.split(',')
        )

    def __parse_to_positive_int(self, __key: str, __value: str) -> int:
        self.__validate_is_digits(__key, __value)
        int_value = int(__value)
        if int_value < 1:
            raise BadQueryArgError(
                __key,
                __value,
                message=f'The {__key} must be 1 or greater.'
            )
        return int_value

    def __validate_is_digits(self, __key: str, __value: str) -> None:
        if not __value.isdigit():
            raise BadQueryArgError(
                __key,
                __value,
                message=f'The {__key} must be a positive integer.'
            )

    def __parse_to_sort_by_string(self, __key: str, __value: str) -> str:
        if not re.match(r'-?[a-z]', __value):
            raise BadQueryArgError(
                __key,
                __value,
                message=f'The {__key} must be a column name, with or without leading -.'
            )
        return __value

    def __parse_to_boolean(self, __key: str, __value: str) -> bool:
        match __value.lower():
            case 'true':
                return True
            case 'false':
                return False
            case _:
                raise BadQueryArgError(
                    __key,
                    __value,
                    message=f"The {__key} must be either 'True' or 'False'"
                )



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from minio.datatypes import Object as MinioObject

from .parser import Parser
from ..core import DataObject


S3Object = MinioObject
S3Transfer = Iterable[S3Object]


class S3Converter():
    """
    Converts from S3 Minio objects to instances of
    `DataObject`.
    """

    def __init__(
        self,
        parser: Parser
    ) -> None:
        self.__parser = parser

    def convert(self, input_: S3Object) -> DataObject:
        """
        Converts a single S3Object to a DataObject
        """
        return self.__parser.parse(input_)

    def convert_list(
        self,
        input_: S3Transfer
    ) -> Iterable[DataObject]:
        """
        Converts an S3Transfer containing a list of results
        """
        # S3 buckets can be quite big, so we want to make sure that
        # we're staying in generator land
        for obj in input_:
            yield self.__parser.parse(obj)



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from typing import Any, Iterable, Optional

from ..core import DataObject

if typing.TYPE_CHECKING:
    from ..core import DataSource


GoatApiResource = dict[str, Any]
GoatApiDoc = dict[str, list[GoatApiResource]]


class Parser(ABC):
    """
    Parses GoaT API transfer resource `dict`s to `DataObject`
    instances
    """

    def parse_iterable(
        self,
        transfers: Iterable[GoatApiResource]
    ) -> Iterable[DataObject]:
        """
        Parses an `Iterable` of GoaT API transfer resources
        """

        return (
            self.parse(t) for t in transfers
        )

    @abstractmethod
    def parse(self, transfer: GoatApiResource) -> DataObject:
        """
        Parses an individual GoaT transfer resource to a
        `DataObject` instance
        """


class DefaultParser(Parser):

    def __init__(self, data_source_dict: dict[str, DataSource]) -> None:
        self.__dict = data_source_dict

    def parse(self, transfer: GoatApiResource) -> DataObject:
        type_ = 'taxon'
        ds = self.__get_data_source(type_)
        raw_attributes = transfer.get('result')
        attributes = self.__convert_attributes(type_, raw_attributes)

        return ds.data_object_factory(
            type_,
            id_=transfer.get('result').get('taxon_id'),
            attributes=attributes,
            to_one=self.__parse_to_ones(transfer)
        )

    def __get_data_source(self, type_: str) -> DataSource:
        return self.__dict[type_]

    def __convert_attributes(
            self,
            type_: str,
            attributes: Optional[dict[str, Any]]
    ) -> dict[str, Any]:
        ret = {}
        if attributes is None:
            return ret
        # Direct attributes
        for att in ['scientific_name', 'taxon_rank']:
            ret[att] = attributes.get(att)

        # Attributes with values in fields
        normal_fields = [
            'genome_size', 'chromosome_number',
            'haploid_number', 'ploidy', 'assembly_level'
        ]
        one_or_list_fields = ['echabs92', 'habreg_2017', 'marhabreg-2017', 'waca_1981',
                              'isb_wildlife_act_1976', 'protection_of_badgers_act_1992',
                              'family_representative', 'long_list', 'sample_collected',
                              'country_list']
        if 'fields' in attributes:
            for att in normal_fields + one_or_list_fields:
                if att in attributes['fields']:
                    att_value = attributes['fields'][att]
                    if att_value is not None and 'value' in att_value:
                        if att in one_or_list_fields and \
                                not isinstance(att_value['value'], list):
                            ret[att] = [att_value['value']]
                        else:
                            ret[att] = att_value['value']
        if 'names' in attributes:
            for att in attributes['names']:
                if att == 'synonym':
                    synonym_value = attributes['names'][att]['name']
                    if isinstance(synonym_value, list):
                        ret[att] = synonym_value
                    else:
                        ret[att] = [synonym_value]
                else:
                    ret[att] = attributes['names'][att]['name'][0]

        # Lineage
        if 'ranks' in attributes:
            ret['lineage'] = []
            for _, ancestor in attributes['ranks'].items():
                ret['lineage'] = [ancestor['scientific_name']] + ret['lineage']

        return ret

    def __parse_to_ones(
        self,
        transfer: GoatApiResource
    ) -> dict[str, DataObject]:
        raw_attributes = transfer.get('result')
        return {
            k: self.parse({'result': v})
            for k, v in raw_attributes.get('ranks', {}).items()
        }



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC
from collections.abc import Callable, Iterable
from typing import Any, Protocol

from .data_object import DataDict, DataObject
from .data_source_dict import DataSourceDict
from .datasource_error import DataSourceError, NotRelationalError
from .operator import Relational
from .relationship import RelationshipConfig, ToManyDict, ToOneDict

if typing.TYPE_CHECKING:
    from .datasource import DataSource


ToOne = dict[str, DataObject | None]
ToMany = dict[str, Iterable[DataObject]]


class DataObjectFactory(Protocol):
    """
    A factory that takes several args and kwargs, and returns
    a new DataObject instance.
    """

    def __call__(
        self,
        type_: str,

        id_: str | None = None,
        attributes: dict[str, Any] | None = None,
        to_one: ToOne | None = None,
        to_many: ToMany | None = None
    ) -> DataObject:
        ...


class DataSourceDictFactory(Protocol):
    """
    A factory that takes a variable number of `DataSource` instances,
    and returns a `dict` mapping `DataObject` type to its hosting
    `DataSource`.
    """

    def __call__(
        self,
        *data_sources: DataSource
    ) -> dict[str, DataSource]:
        ...


OneDictFactory = Callable[
    [DataObject],
    dict[str, DataObject | None]
]
"""
Takes a `DataObject` instance, returns a `dict` mapping
to-one relationship names to its optional `DataObject`
"""


ManyDictFactory = Callable[
    [DataObject],
    dict[str, Iterable[DataObject]]
]
"""
Takes a `DataObject` instance, returns a `dict` mapping
to-many relationship names to its `Iterable[DataObject]`
"""


def _local_name(__name: str) -> bool:
    __PROPERTY_NAMES = [  # noqa N806
        'id',
        'type',
        'provenance',
        'attributes',
        'to_one_relationships',
        'to_many_relationships',
        'get_field_by_name'
    ]

    return (
        __name.startswith('_')
        or __name in __PROPERTY_NAMES
    )


def core_data_object(
    *data_sources: DataSource,
    one_dict_factory: OneDictFactory = lambda o: ToOneDict(o),
    many_dict_factory: ManyDictFactory = lambda o: ToManyDict(o),
    data_source_dict_factory: DataSourceDictFactory = lambda *d: DataSourceDict(*d)
) -> type[DataObject]:
    """
    Takes a tuple of DataSource instances, and creates a CoreDataObject
    implementation that refers to all of them.

    This must be called for the given DataSources to be able to create
    CoreDataObject instances (as it injects a factory).
    """

    data_source_dict = data_source_dict_factory(*data_sources)

    class CoreDataObject(DataObject, ABC):
        """
        A DataObject that can be created outside of a DataSource, which
        should prove sufficient for most use-cases. Simply set values
        on an instance, and they will be interpreted as either:

        - to-one relationships  - if the value is a single DataObject
        - attributes            - otherwise

        Note there are other supported DataObject classes - any class
        that inherits from DataObject meets the criteria.
        """

        def __init__(
            self,
            type_: str,
            id_: str | None = None,
            provenance_: str | None = None,
            attributes: DataDict | None = None,
            to_one: ToOne | None = None,
            to_many: ToMany | None = None,
            stub: bool = False,
            stub_types: Iterable[str] | None = None
        ):
            self.__id = id_
            self.__type = type_
            self.__provenance = provenance_
            self.__attributes = {} if attributes is None else attributes
            self.__to_one_objects = {} if to_one is None else to_one
            self.__to_many_objects = {} if to_many is None else to_many
            if stub and id_ is None:
                raise DataSourceError('ID must be set if stub is True')
            self.__stub_value = stub
            self.__stub_types = stub_types
            if self.__relational:
                self.__to_one_relations = one_dict_factory(self)
                self.__to_many_relations = many_dict_factory(self)

        def __str__(self) -> str:
            dump = f'type="{self.type}"'

            if self.id is not None:
                dump += f', id="{self.id}"'

            return f'CoreDataObject({dump})'

        def __getattribute__(self, name: str, /) -> Any:
            if _local_name(name):
                return object.__getattribute__(self, name)

            if self.__stub_value:
                self.__unstub()

            if name in self.__to_one_names:
                if name in self._to_one_objects:
                    return self._to_one_objects[name]
                return self.to_one_relationships.get(name)

            if name in self.__to_many_names:
                if name in self._to_many_objects:
                    return self._to_many_objects[name]
                return self.to_many_relationships.get(name, [])

            return self.__attributes.get(name)

        def __setattr__(self, name: str, value: Any, /) -> None:
            if _local_name(name):
                object.__setattr__(self, name, value)
            elif name in self.__to_one_names:
                self._to_one_objects[name] = value
            elif name in self.__to_many_names:
                self._to_many_objects[name] = value
            else:
                self.__attributes[name] = value

        def __unstub(self) -> None:
            self.__stub_value = False
            # Actually get this object from the datasource
            possible_types = self.__stub_types if self.__stub_types is not None else [self.__type]
            for t in possible_types:
                object_from_datasource = data_source_dict[t].get_one(t, self.__id)
                if object_from_datasource is not None:
                    self.__type = object_from_datasource.type
                    self.__attributes = object_from_datasource.attributes
                    self.__to_one_objects = object_from_datasource._to_one_objects
                    break

        @property
        def type(self) -> str:  # noqa
            if self.__stub_value and self.__type is None:
                self.__unstub()
            return self.__type

        @property
        def id(self) -> str | None:  # noqa
            return self.__id

        @id.setter
        def id(self, new_id: str) -> None:  # noqa
            self.__id = new_id

        @property
        def provenance(self) -> str | None:
            return self.__provenance

        @provenance.setter
        def provenance(self, new_provenance: str | None) -> None:
            self.__provenance = new_provenance

        @property
        def attributes(self) -> dict[str, Any]:
            if self.__stub_value:
                self.__unstub()
            return self.__attributes

        @property
        def to_one_relationships(self) -> dict[str, DataObject | None]:
            if not self.__relational:
                raise NotRelationalError(self)
            return self.__to_one_relations

        @property
        def to_many_relationships(self) -> dict[str, list[DataObject]]:
            if not self.__relational:
                raise NotRelationalError(self)
            return self.__to_many_relations

        @property
        def _to_one_objects(self) -> dict[str, DataObject | None]:
            return self.__to_one_objects

        @property
        def _to_many_objects(self) -> dict[str, list[DataObject]]:
            return self.__to_many_objects

        @property
        def __relational(self) -> bool:
            """Whether the hosting DataSource is relational or not"""

            return isinstance(self._host, Relational)

        @property
        def __relationship_config(self) -> RelationshipConfig | None:
            return self._host.relationship_config.get(self.type)

        @property
        def __to_one_names(self) -> list[str]:
            if not self.__relational:
                return []
            cfg = self.__relationship_config
            return (
                [] if cfg is None or cfg.to_one is None
                else list(cfg.to_one.keys())
            )

        @property
        def __to_many_names(self) -> list[str]:
            if not self.__relational:
                return []
            cfg = self.__relationship_config
            return (
                [] if cfg is None or cfg.to_many is None
                else list(cfg.to_many.keys())
            )

        @property
        def _host(self) -> DataSource | Relational:
            return data_source_dict[self.type]

    def core_data_object_factory(
        type_: str,
        id_: str | None = None,
        attributes: dict[str, Any] | None = None,
        to_one: ToOne | None = None,
        to_many: ToMany | None = None,
        stub: bool = False,  # Set stub if only type and id are given on creation
        stub_types: Iterable[str] | None = None
    ) -> DataObject:

        return CoreDataObject(
            type_,
            id_=id_,
            attributes=attributes,
            to_one=to_one,
            to_many=to_many,
            stub=stub,
            stub_types=stub_types
        )

    for ds in data_sources:
        ds.data_object_factory = core_data_object_factory

    return CoreDataObject



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


from ..core import (
    core_data_object
)
from ..gap import (
    GapDataSource
)


def gap(**kwargs) -> GapDataSource:
    gap = GapDataSource({
        'uri': 's3://gap/data/assembly.json',
        'type': 'assembly',
        'id_attribute': 'accession',
        's3_host': 'cog.sanger.ac.uk',
        's3_access_key': None,
        's3_secret_key': None,
        'mappings': {
            'project': {'heading': 'project', 'type': 'str'},
            'phylum': {'heading': 'phylum', 'type': 'str'},
            'species': {'heading': 'species', 'type': 'str'},
            'accession': {'heading': 'accession', 'type': 'str'},
            'assembly_name': {'heading': 'assembly_name', 'type': 'str'},
            'results': {'heading': 'results', 'type': 'str'},
            'taxon_id': {'heading': 'taxon_id', 'type': 'int'},
            'phylum_id': {'heading': 'phylum_id', 'type': 'str'},
            'image_url': {'heading': 'image_url', 'type': 'str'},
            'image_caption': {'heading': 'image_caption', 'type': 'str'},
            'lustre_path_analysis_base': {
                'heading': 'lustre_path_analysis_base', 'type': 'str'
            },
            'lustre_path_assembly': {
                'heading': 'lustre_path_assembly', 'type': 'str'
            },
            'lustre_path_species': {
                'heading': 'lustre_path_species', 'type': 'str'
            }
        }
    })
    core_data_object(gap)
    return gap



# SPDX-FileCopyrightText: 2026 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Any

from . import ElasticDataSource
from .client import ElasticClient
from .converter import (
    ElasticApiConverter,
    ElasticUpdateInputConverter,
    ElasticUpsertInputConverter,
)
from .parser import (
    DefaultElasticApiParser,
    DefaultElasticUpdateInputParser,
    DefaultElasticUpsertInputParser,
)
from ..core import (
    AttributeMetadata,
    DefaultAttributeMetadata
)
from ..core.relationship import RelationshipConfig


class _ConverterFactoriesManager:
    """
    The purpose of this class is to provide `elastic_api_converter_factory` and
    `data_object_converter_factory`, functions passed to `ElasticDataSource` that manage the
    instantiation of their respective converters (`ElasticApiConverter`, `DataObjectConverter`
    and `DataObjectUpdateConverter`).
    The reason we need this manager class around these factory functions is because the parser
    classes (needed to instantiate the converters) cannot themselves be instantiated until we
    already have an instance of `ElasticDataSource`.
    So, this class is instantiated, then its converter factory methods are provided to the data
    source for its instantiation, then the `data_source` property is set to enable the factory
    functions to work.
    """
    __slots__ = ['__data_source']
    __data_source: ElasticDataSource | None

    def __init__(self) -> None:
        # The factory is instantisated before the data source, so this must be assigned
        # after initialisation. Therefore, if `None`, the data source hasn't been instantiated yet
        self.__data_source = None

    @property
    def data_source(self) -> ElasticDataSource | None:
        """
        Fetch the data source, or `None` if it hasn't been instantiated yet
        """
        return self.__data_source

    @data_source.setter
    def data_source(self, data_source: ElasticDataSource) -> None:
        """
        The means to link the data source once it has been instantiated
        """
        self.__data_source = data_source

    def elastic_api_converter_factory(self) -> ElasticApiConverter:
        if self.data_source is None:
            raise Exception(
                'TOL INTERNAL ERROR: factory function for ElasticApiConverter called '
                'before the data source was assigned in _ConverterFactoriesManager'
            )

        parser = DefaultElasticApiParser(self.data_source)
        return ElasticApiConverter(parser)

    def data_object_converter_factory(self) -> ElasticUpsertInputConverter:
        if self.data_source is None:
            raise Exception(
                'TOL INTERNAL ERROR: factory function for DataObjectConverter called '
                'before the data source was assigned in _ConverterFactoriesManager'
            )

        parser = DefaultElasticUpsertInputParser(self.data_source)
        return ElasticUpsertInputConverter(parser)

    def data_object_update_converter_factory(self) -> ElasticUpdateInputConverter:
        if self.data_source is None:
            raise Exception(
                'TOL INTERNAL ERROR: factory function for DataObjectUpdateConverter called '
                'before the data source was assigned in _ConverterFactoriesManager'
            )

        parser = DefaultElasticUpdateInputParser(self.data_source)
        return ElasticUpdateInputConverter(parser)


def _client_factory() -> ElasticClient:
    """
    A resonable default for creating
    an `ElasticApiClient` instance
    """
    # TODO
    raise NotImplementedError(
        "`ElasticClient` hasn't been made yet, so this function cannot be implemented, "
        'but it is useful in the construction of `ElasticDataSource` to have this function present'
    )


def create_elastic_datasource(
    config: dict,
    attribute_metadata: AttributeMetadata = DefaultAttributeMetadata,
    relationship_cfg: dict[str, RelationshipConfig] | None = None,
    runtime_fields: dict[str, Any] = {},
    **kwargs,
) -> ElasticDataSource:
    """
    Properly instaniates an ElasticDataSource
    using the configuration required for the client
    """
    # Instantiate the factories needed by the data source
    client_factory = _client_factory

    # The converters are different however, as they require a references to the data source itself.
    # Thus, a manager object is used to pass the factory functions to the data source's
    # constructor before its reference is passed in
    converter_factories_manager = _ConverterFactoriesManager()

    # Instantiate the data source
    elastic_ds = ElasticDataSource(
        config,
        client_factory,
        converter_factories_manager.elastic_api_converter_factory,
        converter_factories_manager.data_object_converter_factory,
        converter_factories_manager.data_object_update_converter_factory,
        attribute_metadata,
        relationship_cfg,
        runtime_fields,
        **kwargs,
    )

    # Update the converter factories manager so that the converter factories have a references to
    # the now instantiated data source
    converter_factories_manager.data_source = elastic_ds

    return elastic_ds



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

class Defaults:
    PORTAL_URL = 'https://portal.tol.sanger.ac.uk'
    PORTAL_API_PATH = '/api/v1'
    PORTAL_API_DATA_PATH = '/data'

    STS_URL = 'https://sts.tol.sanger.ac.uk'
    STS_API_PATH = '/api/v2'
    STS_API_DATA_PATH = '/data'

    TOLID_URL = 'https://id.tol.sanger.ac.uk'
    TOLID_API_PATH = '/api/v3'
    TOLID_API_DATA_PATH = ''

    TOLQC_URL = 'https://qc.tol.sanger.ac.uk'
    TOLQC_API_PATH = '/api/v1'
    TOLQC_API_DATA_PATH = '/data'

    TREEOFSEX_URL = 'https://treeofsex.sanger.ac.uk'
    TREEOFSEX_API_PATH = '/api/v1'
    TREEOFSEX_API_DATA_PATH = '/data'

    BIOSCAN_URL = 'https://bioscan.tol.sanger.ac.uk'
    BIOSCAN_API_PATH = '/api/v1'
    BIOSCAN_API_DATA_PATH = ''

    BIOSCAN_IMAGE_BUCKET_NAME = 'tol-bioscan-images'
    BIOSCAN_IMAGE_PREFIX = 'processed_images/'

    WORKFLOWS_URL = 'https://workflows.tol.sanger.ac.uk'
    WORKFLOWS_API_PATH = '/api/v2'
    WORKFLOWS_API_DATA_PATH = '/data'

    GOAT_URL = 'https://goat.genomehubs.org'
    GOAT_API_PATH = '/api/v2'

    BOLD_URL = 'https://data.boldsystems.org'
    BOLD_PORTAL_URL = 'https://portal.boldsystems.org'
    BOLD_API_PATH = '/api'

    COPO_URL = 'https://copo-project.org'
    COPO_API_PATH = '/api'

    JIRA_URL = 'https://jira.sanger.ac.uk'

    LABWHERE_URL = 'https://labwhere.psd.sanger.ac.uk'
    LABWHERE_API_PATH = '/api'

    ENA_URL = 'https://www.ebi.ac.uk'



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections.abc import Iterable as IterableABC
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Iterator

from sqlalchemy import ForeignKey
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import (
    Mapped,
    declared_attr,
    mapped_column,
    relationship
)

from ..model import Model


@dataclass(frozen=True, kw_only=True)
class StandardModels(IterableABC[type[Model]]):
    """
    Contains the needed models for dynamic, user-configurable
    dashboards.

    Additionally contains `_user_mixin`, from which developers
    should inherit their `User` class from `tol.sql.auth`.
    """

    loader: type[Model]
    data_source_instance: type[Model]
    loader_instance: type[Model]
    data_source_config: type[Model]
    data_source_config_attribute: type[Model]
    data_source_config_relationship: type[Model]
    data_source_config_summary: type[Model]
    component: type[Model]
    component_zone: type[Model]
    zone: type[Model]
    zone_view: type[Model]
    view: type[Model]
    view_board: type[Model]
    board: type[Model]
    web_app: type[Model]

    _user_mixin: type[Any]

    def __iter__(self) -> Iterator[type[Model]]:
        """
        Returns in order they should be deleted
        """

        return iter(
            [
                self.component_zone,
                self.component,
                self.zone_view,
                self.zone,
                self.view_board,
                self.view,
                self.board,
                self.data_source_config_summary,
                self.data_source_config_relationship,
                self.data_source_config_attribute,
                self.loader_instance,
                self.data_source_instance,
                self.data_source_config,
                self.loader,
                self.web_app
            ]
        )


def create_standard_models(
    base_model_class: type[Model],
    user_model_class_name: str = 'User',
    user_table_name: str = 'user'
) -> StandardModels:
    """
    Creates all needed models (and joining tables) for
    user-configurable dashboards.

    Returns a `StandardModels` instance that functions like an
    `Iterable`.
    """

    class LoaderInstance(base_model_class):
        __tablename__ = 'loader_instance'

        id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)  # noqa A003

        # Frequency of loading
        frequency_weekly: Mapped[bool] = mapped_column(nullable=True)
        frequency_daily: Mapped[bool] = mapped_column(nullable=True)
        frequency_hourly: Mapped[bool] = mapped_column(nullable=True)
        frequency_quarter_hourly: Mapped[bool] = mapped_column(nullable=True)
        date_last_run: Mapped[datetime] = mapped_column(nullable=True)

        # Loader
        loader_id: Mapped[int] = mapped_column(
            ForeignKey('loader.id'),
            nullable=False
        )
        loader: Mapped['Loader'] = relationship(  # noqa F821
            back_populates='loader_instances',
            foreign_keys=[loader_id]
        )

        # Relationships
        source_data_source_instance_id: Mapped[str] = mapped_column(
            ForeignKey('data_source_instance.id'),
            nullable=False
        )
        source_data_source_instance: Mapped['DataSourceInstance'] = relationship(  # noqa F821
            back_populates='source_loader_instances',
            foreign_keys=[source_data_source_instance_id]
        )

        destination_data_source_instance_id: Mapped[str] = mapped_column(
            ForeignKey('data_source_instance.id'),
            nullable=False
        )
        destination_data_source_instance: Mapped['DataSourceInstance'] = relationship(  # noqa F821
            back_populates='destination_loader_instances',
            foreign_keys=[destination_data_source_instance_id]
        )

        ids_data_source_instance_id: Mapped[str] = mapped_column(
            ForeignKey('data_source_instance.id'),
            nullable=True
        )
        ids_data_source_instance: Mapped['DataSourceInstance'] = relationship(  # noqa F821
            back_populates='ids_loader_instances',
            foreign_keys=[ids_data_source_instance_id]
        )

    class Loader(base_model_class):
        __tablename__ = 'loader'

        id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)  # noqa A003

        source_object_type: Mapped[str] = mapped_column(nullable=False)
        destination_object_type: Mapped[str] = mapped_column(nullable=False)

        object_filters: Mapped[dict] = mapped_column(
            JSONB,
            nullable=True
        )

        provenance_override: Mapped[str] = mapped_column(nullable=False, default='')
        convert_class: Mapped[str] = mapped_column(nullable=True)
        candidate_key: Mapped[dict] = mapped_column(JSONB, nullable=True)

        # For loading by IDs
        ids_object_type: Mapped[str] = mapped_column(nullable=True)
        ids_attribute: Mapped[str] = mapped_column(nullable=True)
        ids_object_filters: Mapped[dict] = mapped_column(JSONB, nullable=True)
        ids_sort_by: Mapped[str] = mapped_column(nullable=True)
        ids_attribute_in_source: Mapped[str] = mapped_column(nullable=True)

        loader_instances: Mapped[list['LoaderInstance']] = relationship(  # noqa F821
            back_populates='loader',
            foreign_keys=[LoaderInstance.loader_id]
        )

    class Component(base_model_class):
        __tablename__ = 'component'

        id: Mapped[str] = mapped_column(  # noqa A003
            primary_key=True
        )

        title: Mapped[str] = mapped_column(nullable=False)
        object_type: Mapped[str] = mapped_column(nullable=False)
        component_type: Mapped[str] = mapped_column(nullable=False)
        widget_type: Mapped[str] = mapped_column(nullable=False)
        config = mapped_column(type_=JSONB(), nullable=False)
        filter = mapped_column(  # noqa A003
            type_=JSONB(),
            nullable=False,
            default={},
            server_default='{}'  # noqa P103
        )
        filter_pass_through: Mapped[bool] = mapped_column(nullable=False)

        user_id: Mapped[int] = mapped_column(
            ForeignKey(f'{user_table_name}.id'),
            nullable=False
        )
        user = relationship(
            user_model_class_name,
            back_populates='components',
            foreign_keys=[user_id]
        )

        data_source_instance_id: Mapped[str] = mapped_column(
            ForeignKey('data_source_instance.id'),
            nullable=False
        )
        data_source_instance = relationship(
            'DataSourceInstance',
            back_populates='components',
            foreign_keys=[data_source_instance_id]
        )

        component_zones: Mapped[list[ComponentZone]] = relationship(
            back_populates='component'
        )

    class ComponentZone(base_model_class):
        __tablename__ = 'component_zone'

        id: Mapped[int] = mapped_column(  # noqa A003
            primary_key=True,
            autoincrement=True
        )

        order: Mapped[int] = mapped_column(nullable=False)

        component_id: Mapped[str] = mapped_column(
            ForeignKey('component.id'),
            nullable=False
        )
        component = relationship(
            'Component',
            back_populates='component_zones',
            foreign_keys=[component_id]
        )

        zone_id: Mapped[str] = mapped_column(
            ForeignKey('zone.id'),
            nullable=False
        )
        zone = relationship(
            'Zone',
            back_populates='component_zones',
            foreign_keys=[zone_id]
        )

    class Zone(base_model_class):
        __tablename__ = 'zone'

        id: Mapped[str] = mapped_column(  # noqa A003
            primary_key=True
        )

        title: Mapped[str] = mapped_column(nullable=False)
        object_type: Mapped[str] = mapped_column(nullable=False)
        filter = mapped_column(  # noqa A003
            type_=JSONB(),
            nullable=False,
            default={},
            server_default='{}'  # noqa P103
        )

        component_zones: Mapped[list[ComponentZone]] = relationship(
            back_populates='zone'
        )

        zone_views: Mapped[list[ZoneView]] = relationship(
            back_populates='zone'
        )

        user_id: Mapped[int] = mapped_column(
            ForeignKey(f'{user_table_name}.id'),
            nullable=False
        )
        user = relationship(
            user_model_class_name,
            back_populates='zones',
            foreign_keys=[user_id]
        )

        data_source_instance_id: Mapped[str] = mapped_column(
            ForeignKey('data_source_instance.id'),
            nullable=False
        )
        data_source_instance = relationship(
            'DataSourceInstance',
            back_populates='zones',
            foreign_keys=[data_source_instance_id]
        )

    class ZoneView(base_model_class):
        __tablename__ = 'zone_view'

        id: Mapped[int] = mapped_column(  # noqa A003
            primary_key=True,
            autoincrement=True
        )

        order: Mapped[int] = mapped_column(nullable=False)

        zone_id: Mapped[str] = mapped_column(
            ForeignKey('zone.id'),
            nullable=False
        )
        zone = relationship(
            'Zone',
            back_populates='zone_views',
            foreign_keys=[zone_id]
        )

        view_id: Mapped[str] = mapped_column(
            ForeignKey('view.id'),
            nullable=False
        )
        view = relationship(
            'View',
            back_populates='zone_views',
            foreign_keys=[view_id]
        )

    class View(base_model_class):
        __tablename__ = 'view'

        id: Mapped[str] = mapped_column(  # noqa A003
            primary_key=True
        )

        title: Mapped[str] = mapped_column(nullable=False)
        filter = mapped_column(  # noqa A003
            type_=JSONB(),
            nullable=False,
            default={},
            server_default='{}'  # noqa P103
        )

        zone_views: Mapped[list[ZoneView]] = relationship(
            back_populates='view'
        )

        view_boards: Mapped[list[ViewBoard]] = relationship(
            back_populates='view'
        )

        user_id: Mapped[int] = mapped_column(
            ForeignKey(f'{user_table_name}.id'),
            nullable=False
        )
        user = relationship(
            user_model_class_name,
            back_populates='views',
            foreign_keys=[user_id]
        )

    class ViewBoard(base_model_class):
        __tablename__ = 'view_board'

        id: Mapped[int] = mapped_column(  # noqa A003
            primary_key=True,
            autoincrement=True
        )

        order: Mapped[int] = mapped_column(nullable=False)

        view_id: Mapped[str] = mapped_column(
            ForeignKey('view.id'),
            nullable=False
        )
        view = relationship(
            'View',
            back_populates='view_boards',
            foreign_keys=[view_id]
        )

        board_id: Mapped[str] = mapped_column(
            ForeignKey('board.id'),
            nullable=False
        )
        board = relationship(
            'Board',
            back_populates='view_boards',
            foreign_keys=[board_id]
        )

    class Board(base_model_class):
        __tablename__ = 'board'

        id: Mapped[str] = mapped_column(  # noqa A003
            primary_key=True
        )

        title: Mapped[str] = mapped_column(nullable=False)
        filter = mapped_column(  # noqa A003
            type_=JSONB(),
            nullable=False,
            default={},
            server_default='{}'  # noqa P103
        )

        view_boards: Mapped[list[ViewBoard]] = relationship(
            back_populates='board'
        )

        user_id: Mapped[int] = mapped_column(
            ForeignKey(f'{user_table_name}.id'),
            nullable=False
        )
        user = relationship(
            user_model_class_name,
            back_populates='boards',
            foreign_keys=[user_id]
        )

    class DataSourceInstance(base_model_class):
        __tablename__ = 'data_source_instance'

        id: Mapped[str] = mapped_column(primary_key=True)  # noqa A003

        builtin_name: Mapped[str] = mapped_column(nullable=False)
        kwargs: Mapped[dict] = mapped_column(JSONB, nullable=True)
        publish: Mapped[bool] = mapped_column(nullable=False, default=False)
        ui_api_details: Mapped[dict] = mapped_column(JSONB, nullable=True)

        data_source_config_id: Mapped[int] = mapped_column(
            ForeignKey('data_source_config.id'),
            nullable=False
        )
        data_source_config: Mapped['DataSourceConfig'] = relationship(  # noqa F821
            back_populates='data_source_instances',
            foreign_keys=[data_source_config_id]
        )

        source_loader_instances: Mapped[list['LoaderInstance']] = relationship(  # noqa F821
            back_populates='source_data_source_instance',
            foreign_keys=[LoaderInstance.source_data_source_instance_id]
        )
        destination_loader_instances: Mapped[list['LoaderInstance']] = relationship(  # noqa F821
            back_populates='destination_data_source_instance',
            foreign_keys=[LoaderInstance.destination_data_source_instance_id]
        )
        ids_loader_instances: Mapped[list['LoaderInstance']] = relationship(  # noqa F821
            back_populates='ids_data_source_instance',
            foreign_keys=[LoaderInstance.ids_data_source_instance_id]
        )
        components: Mapped[list['Component']] = relationship(  # noqa F821
            back_populates='data_source_instance',
            foreign_keys=[Component.data_source_instance_id]
        )
        zones: Mapped[list['Zone']] = relationship(  # noqa F821
            back_populates='data_source_instance',
            foreign_keys=[Zone.data_source_instance_id]
        )

    class DataSourceConfigAttribute(base_model_class):
        __tablename__ = 'data_source_config_attribute'

        id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)  # noqa A003
        data_source_config_id: Mapped[int] = mapped_column(
            ForeignKey('data_source_config.id'),
            nullable=False
        )
        data_source_config: Mapped['DataSourceConfig'] = relationship(  # noqa F821
            back_populates='data_source_config_attributes',
            foreign_keys=[data_source_config_id]
        )

        name: Mapped[str] = mapped_column(nullable=False)
        object_type: Mapped[str] = mapped_column(nullable=False)
        display_name: Mapped[str] = mapped_column(nullable=True)
        description: Mapped[str] = mapped_column(nullable=True)
        available_on_relationships: Mapped[bool] = mapped_column(nullable=False, default=True)
        is_authoritative: Mapped[bool] = mapped_column(nullable=False, default=False)
        source: Mapped[str] = mapped_column(nullable=True)
        acts_as: Mapped[str] = mapped_column(nullable=True)
        runtime_definition: Mapped[dict] = mapped_column(JSONB, nullable=True)

    class DataSourceConfigRelationship(base_model_class):
        __tablename__ = 'data_source_config_relationship'

        id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)  # noqa A003
        object_type: Mapped[str] = mapped_column(nullable=False)
        name: Mapped[str] = mapped_column(nullable=False)
        foreign_object_type: Mapped[str] = mapped_column(nullable=False)
        foreign_name: Mapped[str] = mapped_column(nullable=False)

        data_source_config_id: Mapped[int] = mapped_column(
            ForeignKey('data_source_config.id'),
            nullable=False
        )
        data_source_config: Mapped['DataSourceConfig'] = relationship(  # noqa F821
            back_populates='data_source_config_relationships',
            foreign_keys=[data_source_config_id]
        )

        source_order: Mapped[list[str]] = mapped_column(JSONB, nullable=True)

    class DataSourceConfigSummary(base_model_class):
        __tablename__ = 'data_source_config_summary'

        id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)  # noqa A003

        source_object_type: Mapped[str] = mapped_column(nullable=False)
        destination_object_type: Mapped[str] = mapped_column(nullable=True)

        object_filters: Mapped[dict] = mapped_column(
            JSONB,
            nullable=False,
            default={}
        )

        group_by: Mapped[dict] = mapped_column(
            JSONB,
            nullable=False,
            default=[]
        )

        stats_fields: Mapped[dict] = mapped_column(
            JSONB,
            nullable=False,
            default=[]
        )

        stats: Mapped[dict] = mapped_column(
            JSONB,
            nullable=False,
            default=[]
        )

        provenance_override: Mapped[str] = mapped_column(
            nullable=False,
            default=''
        )

        data_source_config_id: Mapped[int] = mapped_column(
            ForeignKey('data_source_config.id'),
            nullable=False
        )

        data_source_config: Mapped['DataSourceConfig'] = relationship(  # noqa F821
            back_populates='data_source_config_summaries',
            foreign_keys=[data_source_config_id]
        )

    class DataSourceConfig(base_model_class):
        __tablename__ = 'data_source_config'

        id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)  # noqa A003

        name: Mapped[str] = mapped_column(nullable=False)
        description: Mapped[str] = mapped_column(nullable=False)

        data_source_instances: Mapped[list['DataSourceInstance']] = relationship(  # noqa F821
            back_populates='data_source_config',
            foreign_keys=[DataSourceInstance.data_source_config_id]
        )
        data_source_config_attributes: Mapped[list['DataSourceConfigAttribute']] = relationship(  # noqa F821
            back_populates='data_source_config',
            foreign_keys=[DataSourceConfigAttribute.data_source_config_id]
        )
        data_source_config_relationships: Mapped[list['DataSourceConfigRelationship']] = relationship(  # noqa F821
            back_populates='data_source_config',
            foreign_keys=[DataSourceConfigRelationship.data_source_config_id]
        )
        data_source_config_summaries: Mapped[list['DataSourceConfigSummary']] = relationship(
            back_populates='data_source_config',
            foreign_keys=[DataSourceConfigSummary.data_source_config_id]
        )

    class WebApp(base_model_class):
        __tablename__ = 'web_app'

        id : Mapped[str] = mapped_column(primary_key=True) # noqa A003

        navigation: Mapped[dict] = mapped_column(
            JSONB,
            nullable=False,
            default={},
            server_default='{}'  # noqa P103
        )
        profile_navigation: Mapped[dict] = mapped_column(
            JSONB,
            nullable=False,
            default={},
            server_default='{}'  # noqa P103
        )

    class _UserMixin:

        @declared_attr
        def components(self) -> Mapped[list[Component]]:
            return relationship(
                back_populates='user'
            )

        @declared_attr
        def zones(self) -> Mapped[list[Zone]]:
            return relationship(
                back_populates='user'
            )

        @declared_attr
        def views(self) -> Mapped[list[View]]:
            return relationship(
                back_populates='user'
            )

        @declared_attr
        def boards(self) -> Mapped[list[Board]]:
            return relationship(
                back_populates='user'
            )

    return StandardModels(
        loader=Loader,
        data_source_instance=DataSourceInstance,
        loader_instance=LoaderInstance,
        data_source_config=DataSourceConfig,
        data_source_config_attribute=DataSourceConfigAttribute,
        data_source_config_relationship=DataSourceConfigRelationship,
        data_source_config_summary=DataSourceConfigSummary,
        component=Component,
        component_zone=ComponentZone,
        zone=Zone,
        zone_view=ZoneView,
        view=View,
        view_board=ViewBoard,
        board=Board,
        web_app=WebApp,
        _user_mixin=_UserMixin
    )



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC
from typing import Any, Callable

from .model import Model
from ..core import DataObject, ReqFieldsTree
from ..core.core_converter import Converter
from ..core.factory import DataObjectFactory


TypeFunction = Callable[[Model], str]
"""Takes a Model instance, and returns the corresponding DataObject type."""


class ModelConverter(Converter[Model, DataObject], ABC):
    """
    Converts Sqlalchemy model instances to DataObject instances.
    """


class DefaultModelConverter(ModelConverter):
    def __init__(
        self,
        type_function: TypeFunction,
        data_object_factory: DataObjectFactory,
        requested_tree: ReqFieldsTree,
    ) -> None:
        """
        Takes a type_function Callable, which determines the type of the
        DataObject for a given Model instance.
        """

        self.__type_function = type_function
        self.__data_object_factory = data_object_factory
        self.__requested_tree = requested_tree

    def convert(self, model: Model) -> DataObject:
        if model is None:
            return None
        if tree := self.__requested_tree:
            return self.__convert_requested(model, tree)
        else:
            return self.__data_object_factory(
                self.__type_function(model),
                id_=model.instance_id,
                attributes=model.instance_attributes,
            )

    def __convert_requested(self, model, tree):
        if attr_names := tree.attribute_names:
            attributes = {x: getattr(model, x) for x in attr_names if x != 'id'}
        else:
            attributes = model.instance_attributes

        req_to_ones = self.__convert_to_ones_requested(model, tree)
        req_to_many = self.__convert_to_many_requested(model, tree)

        obj = self.__data_object_factory(
            self.__type_function(model),
            id_=model.instance_id,
            attributes=attributes,
            to_one=req_to_ones,
            to_many=req_to_many,
        )
        return obj

    def __convert_to_ones_requested(self, model, tree):
        to_ones = {}
        for rel_name, remote in model.get_to_one_relationship_config().items():
            one = None
            if sub_tree := tree.get_sub_tree(rel_name):
                if sub_model := getattr(model, rel_name):
                    one = self.__convert_requested(sub_model, sub_tree)
            else:
                # Create a stub DataObject
                rel_col = model.get_foreign_key_name(rel_name)
                if rel_id := getattr(model, rel_col):
                    one = self.__data_object_factory(remote, id_=rel_id)
            to_ones[rel_name] = one
        return to_ones if to_ones else None

    def __convert_to_many_requested(self, model, tree):
        to_manys = {}
        for rel_name in model.get_to_many_relationship_config():
            if sub_tree := tree.get_sub_tree(rel_name):
                to_manys[rel_name] = [
                    self.__convert_requested(x, sub_tree) for x in getattr(model, rel_name)
                ]
        return to_manys if to_manys else None


class DataObjectConverter(Converter[DataObject, Model], ABC):
    """
    Converts `DataObject` instances to `Model` instances.
    """


class DefaultDataObjectConverter(DataObjectConverter):
    def __init__(self, type_models_dict: dict[str, type[Model]]) -> None:
        """
        `type_models_dict` maps object type to the
        corresponding `type[Model]` class.
        """

        self.__models_dict = type_models_dict

    def convert(self, input_: DataObject) -> Model:
        model_class = self.__models_dict[input_.type]

        return model_class(
            **self.__get_id_dict(input_.id, model_class),
            **input_.attributes,
            **self.__get_relation_dict(model_class, input_._to_one_objects),
        )

    def __get_id_dict(self, id_: str, model_class: type[Model]) -> dict[str, str]:
        id_column_name = model_class.get_id_column_name()
        return {id_column_name: id_}

    def __get_relation_dict(
        self, model_class: type[Model], ones: dict[str, DataObject]
    ) -> dict[str, str]:
        # TODO validation - relationship names and their types

        return {
            model_class.get_foreign_key_name(rel_name): self.__map_to_foreign_key(rel_obj)
            for rel_name, rel_obj in ones.items()
        }

    def __map_to_foreign_key(self, rel_obj: DataObject | None) -> Any | None:
        return None if rel_obj is None else rel_obj.id



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)
from ...core.operator.updater import DataObjectUpdate


class ElasticTolidToElasticSampleUpdateConverter(
        DataObjectToDataObjectOrUpdateConverter):

    def convert(self, data_object: DataObject) -> Iterable[DataObjectUpdate]:
        if data_object.tolid_species is not None and data_object.tolid_specimen is not None:
            species = data_object.to_one_relationships['tolid_species']
            specimen = data_object.to_one_relationships['tolid_specimen']
            yield (
                None,
                {
                    'tolid_tolid': self._data_object_factory(
                        'tolid',
                        data_object.id
                    ),
                    'sts_species.id':
                        data_object.requested_taxonomy_id
                        if data_object.requested_taxonomy_id is not None
                        else species.id,
                    'sts_specimen.id': specimen.id
                }
            )



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Iterable

from more_itertools import flatten

from .data_object import DataObject
from .factory import DataObjectFactory
from .operator.updater import DataObjectUpdate


class DataObjectToDataObjectOrUpdateConverter(ABC):
    """
    This is (currently) not inheriting from Converter as we are changing
    the method signatures (convert() can return None, a DataObject or an
    Iterable of DataObjects)
    """

    def __init__(
        self,
        data_object_factory: DataObjectFactory
    ) -> None:
        """
        Takes a data_object_factory to use for creating new DataObjects
        """
        self._data_object_factory = data_object_factory
        self._return_objects = []

    def convert_iterable(
        self,
        inputs: Iterable[DataObject | DataObjectUpdate]
    ) -> Iterable[DataObject]:
        return flatten((self.convert(i) for i in inputs))

    def get_return_objects(self):
        return self._return_objects

    @abstractmethod
    def convert(self, input_: DataObject) -> Iterable[DataObject | DataObjectUpdate]:
        """
        Converts an input representation to an output representation.
        output can be:
            None: ignored
            DataObject
            Iterable of DataObjects
        """


class DefaultDataObjectToDataObjectConverter(DataObjectToDataObjectOrUpdateConverter):

    def __init__(
        self,
        data_object_factory: DataObjectFactory,
        destination_object_type: str | None = None,
        id_field: str | None = None
    ):

        super().__init__(data_object_factory)
        self.__dest_object_type = destination_object_type
        self.__id_field = id_field

    def convert(
        self,
        data_object: DataObject
    ) -> Iterable[DataObject]:
        """
        A "passthrough" converter only dealing with attributes:
            None: ignored
            DataObject
            Iterable of DataObjects
        """
        if data_object is not None and data_object.id is not None:
            dest_type = (
                self.__dest_object_type
                if self.__dest_object_type is not None
                else self.data_loader._destination_object_type
            )
            ret = self._data_object_factory(
                id_=(
                    data_object.id if self.__id_field is None
                    else data_object.get_field_by_name(self.__id_field)
                ),
                type_=dest_type,
                attributes={
                    k: v
                    for k, v in data_object.attributes.items()
                    if k != self.__id_field
                }
            )
            yield ret


class SanitisingConverter(DataObjectToDataObjectOrUpdateConverter):

    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        pass

    def __init__(
        self,
        data_object_factory: DataObjectFactory,
        config: Config,
        **kwargs
    ):
        super().__init__(data_object_factory)

    def convert(
        self,
        data_object: DataObject
    ) -> Iterable[DataObject]:
        """
        A converter that removes leading and trailing whitespace from
        string attributes.
        """
        if data_object is not None and data_object.id is not None:
            ret = self._data_object_factory(
                id_=data_object.id,
                type_=data_object.type,
                attributes={
                    k: self.__sanitise(v)
                    for k, v in data_object.attributes.items()
                }
            )
            yield ret

    def __sanitise(
        self,
        value: object
    ) -> object:
        if isinstance(value, str):
            return value.strip()
        return value



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from .benchling_datasource import BenchlingDataSource  # noqa
from .benchling_warehouse_datasource import BenchlingWarehouseDataSource  # noqa



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os

from ..core import (
    core_data_object
)
from ..sciops import (
    SequencingDataSource
)


def sciops(**kwargs) -> SequencingDataSource:
    sciops = SequencingDataSource({
        'redpanda_url': os.getenv('REDPANDA_URL'),
        'redpanda_api_key': os.getenv('REDPANDA_API_KEY'),
        'rabbitmq_host': os.getenv('RABBITMQ_HOST'),
        'rabbitmq_port': os.getenv('RABBITMQ_PORT'),
        'rabbitmq_username': os.getenv('RABBITMQ_USERNAME'),
        'rabbitmq_password': os.getenv('RABBITMQ_PASSWORD'),
        'rabbitmq_vhost': os.getenv('RABBITMQ_VHOST'),
        'rabbitmq_exchange': os.getenv('RABBITMQ_EXCHANGE'),
        'rabbitmq_routing_key': os.getenv('RABBITMQ_ROUTING_KEY'),
        'rabbitmq_use_ssl': os.getenv('RABBITMQ_USE_SSL'),
        'rabbitmq_publish_retry_delay': os.getenv('RABBITMQ_PUBLISH_RETRY_DELAY'),
        'rabbitmq_publish_retries': os.getenv('RABBITMQ_PUBLISH_RETRIES'),
        'tol_feedback_queue': os.getenv('TOL_FEEDBACK_QUEUE')
    })
    core_data_object(sciops)
    return sciops



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections.abc import Iterable as IterableABC
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Iterator

from sqlalchemy import (
    CheckConstraint,
    ForeignKey,
    UniqueConstraint
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import (
    Mapped,
    declared_attr,
    mapped_column,
    relationship
)

from ..model import Model


@dataclass(frozen=True, kw_only=True)
class ActionModels(IterableABC[type[Model]]):
    """
    Contains the needed models for actions.

    Additionally contains `_user_mixin`, from which developers
    should inherit their `User` class from `tol.sql.auth`.
    """

    action: type[Model]
    user_action: type[Model]
    role_action: type[Model]

    _user_mixin: type[Any]
    _role_mixin: type[Any]

    def __iter__(self) -> Iterator[type[Model]]:
        """
        Returns in order they should be deleted
        """

        return iter(
            [
                self.user_action,
                self.role_action,
                self.action
            ]
        )


def create_action_models(
    base_model_class: type[Model]
) -> ActionModels:
    """
    Creates all needed models (and joining tables) for
    actions.

    Returns a `ActionModels` instance that functions like an
    `Iterable`.
    """

    class Action(base_model_class):
        __tablename__ = 'action'

        __table_args__ = (
            UniqueConstraint('name', 'object_type'),
        )

        id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)  # noqa A003

        name: Mapped[str] = mapped_column(nullable=False)
        object_type: Mapped[str] = mapped_column(nullable=False)

        flow_name: Mapped[str] = mapped_column(nullable=True)
        class_name: Mapped[str] = mapped_column(nullable=True)
        params: Mapped[dict] = mapped_column(
            JSONB,
            nullable=False,
            default={}
        )

        class_name: Mapped[str] = mapped_column(nullable=True)

        user_actions: Mapped[list['UserAction']] = relationship(  # noqa F821
            back_populates='action'
        )

        role_actions: Mapped[list['RoleAction']] = relationship(  # noqa F821
            back_populates='action'
        )

    class UserAction(base_model_class):
        __tablename__ = 'user_action'

        __table_args__ = (
            CheckConstraint('NOT(ids IS NULL AND filters IS NULL)'),
        )

        id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)  # noqa A003

        created_at: Mapped[datetime] = mapped_column(
            nullable=False,
            default=datetime.now
        )
        params: Mapped[dict] = mapped_column(
            JSONB,
            nullable=False,
            default={}
        )

        ids: Mapped[list[str]] = mapped_column(JSONB, nullable=True)
        filters: Mapped[dict[str, Any]] = mapped_column(JSONB, nullable=True)

        action_id: Mapped[int] = mapped_column(
            ForeignKey('action.id'),
            nullable=False
        )
        action: Mapped['Action'] = relationship(  # noqa F821
            back_populates='user_actions',
            foreign_keys=[action_id]
        )

        user_id: Mapped[int] = mapped_column(
            ForeignKey('user.id'),
            nullable=False
        )
        user: Mapped['User'] = relationship(  # noqa F821
            back_populates='user_actions',
            foreign_keys=[user_id]
        )

    class RoleAction(base_model_class):
        __tablename__ = 'role_action'

        id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)  # noqa A003

        action_id: Mapped[int] = mapped_column(
            ForeignKey('action.id'),
            nullable=False
        )
        action: Mapped['Action'] = relationship(  # noqa F821
            back_populates='role_actions',
            foreign_keys=[action_id]
        )

        role_id: Mapped[int] = mapped_column(
            ForeignKey('role.id'),
            nullable=False
        )
        role: Mapped['Role'] = relationship(  # noqa F821
            back_populates='role_actions',
            foreign_keys=[role_id]
        )

    class _UserMixin:

        @declared_attr
        def user_actions(self) -> Mapped[list[UserAction]]:
            return relationship(
                back_populates='user'
            )

    class _RoleMixin:

        @declared_attr
        def role_actions(self) -> Mapped[list[RoleAction]]:
            return relationship(
                back_populates='role'
            )

    return ActionModels(
        action=Action,
        user_action=UserAction,
        role_action=RoleAction,
        _user_mixin=_UserMixin,
        _role_mixin=_RoleMixin
    )



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os

from .defaults import Defaults
from ..copo import (
    CopoDataSource,
    create_copo_datasource
)
from ..core import (
    core_data_object
)


def copo(**kwargs) -> CopoDataSource:
    copo = create_copo_datasource(
        copo_url=os.getenv('COPO_URL', Defaults.COPO_URL)
        + os.getenv('COPO_API_PATH', Defaults.COPO_API_PATH)
    )
    core_data_object(copo)
    return copo



# SPDX-FileCopyrightText: 2026 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from ..excel import (
    CsvDataSource,
    google_csv_datasource_factory,
)


def google_csv(
    google_csv_id: str,
    object_type: str = 'sheet_row',
    type_mapping: dict | None = None,
    **kwargs,
) -> CsvDataSource:
    return google_csv_datasource_factory(
        google_file_id=google_csv_id,
        object_type=object_type,
        type_mapping=type_mapping,
        **kwargs,
    )



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Iterable


from .data_object import DataObject, ErrorObject


class ValidationSeverity(str, Enum):
    ERROR = 'error'
    WARNING = 'warning'


@dataclass(frozen=True, kw_only=True)
class ValidationResult:
    object_id: str
    """Maps to the row number from the original spreadsheet"""

    detail: str
    """A helpful error message"""

    severity: ValidationSeverity
    """Either `'warning'` or `'error'`"""

    field: str | list[str] | None = None
    """
    The field(s) that failed validation

    `None` indicates the all fields failed on this row
    """

    code: str | None = None
    """An (optional) reference to an external error code"""


class Validator(ABC):
    """
    Validates a stream of `DataObject` instances.

    Note - a `Validator` child does not alter
    said stream.
    """

    def __init__(self) -> None:
        super().__init__()

        self.__results: list[ValidationResult] = []

    @abstractmethod
    def _validate_data_object(
        self,
        obj: DataObject
    ) -> None:
        """Validates a single `DataObject` instance."""

    def validate(
        self,
        object_stream: Iterable[DataObject | ErrorObject]
    ) -> Iterable[DataObject | ErrorObject]:
        """
        Validates a stream of `DataObject` instances.
        """

        for obj in object_stream:
            if isinstance(obj, DataObject):
                self._validate_data_object(obj)

            yield obj

        self._post_validation(obj)

    def add_warning(
        self,
        *,
        object_id: str,
        detail: str,
        field: str | list[str] | None = None,
        code: str | None = None,
    ) -> None:

        self._add_result(
            object_id=object_id,
            detail=detail,
            severity=ValidationSeverity.WARNING,
            field=field,
            code=code,
        )

    def add_error(
        self,
        *,
        object_id: str,
        detail: str,
        field: str | list[str] | None = None,
        code: str | None = None,
    ) -> None:

        self._add_result(
            object_id=object_id,
            detail=detail,
            severity=ValidationSeverity.ERROR,
            field=field,
            code=code,
        )

    @property
    def results(self) -> list[ValidationResult]:
        return self.__results

    @property
    def warnings(self) -> list[ValidationResult]:
        return list(
            self.__get_results_by_severity(
                ValidationSeverity.WARNING
            )
        )

    @property
    def errors(self) -> list[ValidationResult]:
        return list(
            self.__get_results_by_severity(
                ValidationSeverity.ERROR
            )
        )

    @property
    def has_no_errors(self) -> bool:
        """
        Returns `True` if there are no validation errors.
        """

        error_results = self.__get_results_by_severity(
            ValidationSeverity.ERROR,
        )

        return not any(error_results)

    def _post_validation(
        self,
        obj: DataObject,
    ) -> None:
        """
        Override if you need code to run after
        all objects have been validated.
        """

    def _add_result(
        self,
        *,
        object_id: str,
        detail: str,
        severity: ValidationSeverity,
        field: str | list[str] | None = None,
        code: str | None = None,
    ) -> None:

        result = ValidationResult(
            object_id=object_id,
            detail=detail,
            severity=severity,
            field=field,
            code=code,
        )

        self.__results.append(result)

    def __get_results_by_severity(
        self,
        severity: ValidationSeverity
    ) -> Iterable[ValidationResult]:

        return (
            r
            for r in self.__results
            if r.severity == severity
        )



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Any, List

from tol.core import DataObject
from tol.core.validate import Validator


class AllowedValuesValidator(Validator):
    """
    Validates an incoming stream of `DataObject` instances
    according to the specified allowed values for a given
    key.
    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        field: str
        allowed_values: List[Any]
        is_error: bool = True
        detail: str = 'Value is not allowed for the given key'

    __slots__ = ['__config']
    __config: Config

    def __init__(
        self,
        config: Config,
        **kwargs
    ) -> None:

        super().__init__()

        self.__config = config

    def _validate_data_object(
        self,
        obj: DataObject
    ) -> None:

        for key, value in obj.attributes.items():
            if key == self.__config.field and value not in self.__config.allowed_values:
                self.__add_result(obj, key)

    def __add_result(
        self,
        obj: DataObject,
        key: str,
    ) -> None:

        if self.__config.is_error:
            self.add_error(
                object_id=obj.id,
                detail=self.__config.detail,
                field=key
            )
        else:
            self.add_warning(
                object_id=obj.id,
                detail=self.__config.detail,
                field=key,
            )



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod
from datetime import date, datetime
from json import dumps
from typing import Any, Callable, Optional

from ..core import DataSourceFilter


class ApiFilter(ABC):
    """
    Converts instances of `DataSourceFilter` to a
    valid JSON:API filter dump.
    """

    @abstractmethod
    def dumps(self, filter_: DataSourceFilter) -> Optional[str]:
        """
        Emit a filter string from a `DataSourceFilter` instance
        """


def date_serial(v: Any) -> Any:
    if isinstance(v, (date, datetime)):
        return v.isoformat()
    return v


DictDumper = Callable[[dict[str, Any]], str]
default_dict_dumper: DictDumper = lambda d: dumps(
    d,
    separators=(',', ':'),
    default=date_serial
)


class DefaultApiFilter(ApiFilter):

    __KEYS = ['exact', 'contains', 'in_list', 'range', 'and_']

    def __init__(
        self,
        dict_dumper: DictDumper = default_dict_dumper
    ) -> None:

        self.__dict_dumper = dict_dumper

    def dumps(self, filter_: DataSourceFilter) -> Optional[str]:
        __dict = self.__to_dict(filter_)
        return self.__dict_dumper(__dict)

    def __to_dict(self, filter_: DataSourceFilter) -> dict[str, Any]:
        pairs = (
            (k, getattr(filter_, k))
            for k in self.__KEYS
        )

        return {
            k: v for k, v in pairs if v is not None
        }



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import List

from tol.core import DataObject, DataSourceUtils
from tol.core.validate import Validator
from tol.sources.portaldb import portaldb


class AllowedValuesFromDataSourceValidator(Validator):
    """
    Validates that a stream of `DataObject` instances
    contains field that is part of a list.
    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        datasource_instance_id: int
        datasource_object_type: str
        datasource_field_name: str
        field_name: str

    __slots__ = ['__config', '__cached_list']
    __config: Config
    __cached_list: List[str | int | float]

    def __init__(
        self,
        config: Config,
        allowed_values: List[str | int | float] | None = None,  # For testing
        **kwargs
    ) -> None:

        super().__init__()

        self.__config = config
        self.__cached_list = allowed_values \
            or self.__initialize_list_from_datasource()

    def __initialize_list_from_datasource(self) -> List[str | int | float]:
        dsi = portaldb().get_one('data_source_instance', self.__config.datasource_instance_id)
        ds = DataSourceUtils.get_datasource_by_datasource_instance(dsi)
        return [
            obj.get_field_by_name(
                self.__config.datasource_field_name
            ) for obj in ds.get_list(
                self.__config.datasource_object_type
            )
        ]

    def _validate_data_object(
        self,
        obj: DataObject
    ) -> None:
        field_value = obj.get_field_by_name(self.__config.field_name)
        if not field_value:
            return
        if not isinstance(field_value, list):
            field_value = [field_value]
        if any(value not in self.__cached_list for value in field_value):
            multiple_cached_values = len(self.__cached_list) > 1

            cached_list_str = ''
            if multiple_cached_values:
                for index, field in enumerate(self.__cached_list):
                    if index == 0:
                        # First item in the list
                        cached_list_str += f'{field}'
                    elif index == len(self.__cached_list) - 1:
                        # Last item in the list
                        cached_list_str += f' or {field}'
                    else:
                        # Middle items
                        cached_list_str += f', {field}'
            else:  # Only one field
                cached_list_str = self.__cached_list[0]

            # This is extracted rather than being evaluated in the f-string
            # because otherwise the linter doesn't like it!
            pluralisation = ' one of' if multiple_cached_values else ''

            self.add_error(
                object_id=obj.id,
                detail=f'The value of the field {self.__config.field_name} '
                       f'must be{pluralisation} {cached_list_str} '
                       f'(found value {field_value})',
                field=self.__config.field_name,
            )



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod
from typing import Optional

from .mapper import JiraMapper


class JiraSorter(JiraMapper, ABC):
    """Runs order_by against a query"""

    @abstractmethod
    def sort(
        self,
        sort_by: Optional[str]
    ) -> str:
        """Sorts a query using the given models"""


class DefaultJiraSorter(JiraSorter):

    def __init__(
        self,
        field_mappings: dict[str, dict[str, str]]
    ) -> None:
        super().__init__(field_mappings)

    def sort(
        self,
        sort_by: str
    ) -> str:
        if sort_by is None:
            desc = None
            term = 'key'
        elif sort_by.startswith('-'):
            desc = True
            term = sort_by[1:]
        else:
            desc = False
            term = sort_by
        mapped_term = self._map_field(term)
        return f' ORDER BY {mapped_term} {"DESC" if desc else "ASC"}'



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from datetime import timedelta

from prefect import task
from prefect.engine.signals import FAIL

from .logger import get_prefect_logger
from ..eln import (
    generate_assay_results,
    generate_containers,
    generate_custom_entities,
    generate_workflow_outputs,
    generate_workflow_tasks,
    get_benchling_instance,
    sanitise_value
)
from ..sts import (
    get_datetime_setting,
    sts_requests,
    update_datetime_setting,
)


def get_sanger_sample_ids_for_container_list(container_ids, eln_schema_id):
    benchling = get_benchling_instance()
    ret = {}
    # We can only get 20 at once
    for containers_page in [container_ids[i:i + 20] for i in range(0, len(container_ids), 20)]:
        assay_results_page = generate_assay_results(
            benchling,
            schema_id=eln_schema_id,
            storage_ids=containers_page
        )
        for assay_result in assay_results_page:
            container_fluidx_id = sanitise_value(
                assay_result.fields.to_dict()['sample_tube']['displayValue'])
            sanger_sample_id = sanitise_value(
                assay_result.fields.to_dict()['sanger_sample_id']['value'])
            ret[container_fluidx_id] = sanger_sample_id
    get_prefect_logger().info('Found this many Sanger Sample IDs: ' + str(len(ret)))
    return ret


@task(max_retries=3, retry_delay=timedelta(seconds=60))
def add_sanger_sample_ids(submissions, eln_sanger_sample_id_schema_id):
    container_ids = [d['container_eln_id'] for d in submissions]
    sanger_sample_ids = get_sanger_sample_ids_for_container_list(
        container_ids,
        eln_sanger_sample_id_schema_id)
    ret = []
    for submission in submissions:
        if submission['fluidx_id'] in sanger_sample_ids:
            ret.append({**submission,
                        'sanger_sample_id': sanger_sample_ids[submission['fluidx_id']]})
        else:
            get_prefect_logger().warning('Cannot find Sanger Sample ID for tube: '
                                         + submission['fluidx_id'])
    get_prefect_logger().info('Total number of viable submissions: ' + str(len(ret)))
    return ret


def get_fluidx_ids_for_workflow_task_list(workflow_task_ids):
    benchling = get_benchling_instance()
    ret = {}
    # We can only get 20 at once
    for workflow_task_id_page in \
            [workflow_task_ids[i:i + 20] for i in range(0, len(workflow_task_ids), 20)]:
        workflow_tasks_page = generate_workflow_tasks(
            benchling,
            ids=workflow_task_id_page
        )
        for workflow_task in workflow_tasks_page:
            workflow_task_id = workflow_task.id
            fluidx_id = sanitise_value(
                workflow_task.fields.to_dict()['Sample Tube']['displayValue'])
            ret[workflow_task_id] = fluidx_id
    get_prefect_logger().info('Found this many FluidX IDs: ' + str(len(ret)))
    return ret


@task(max_retries=3, retry_delay=timedelta(seconds=60))
def add_fluidx_ids(submissions):
    workflow_task_ids = [d['workflow_task_id'] for d in submissions]
    fluidx_ids = get_fluidx_ids_for_workflow_task_list(
        workflow_task_ids)
    ret = []
    for submission in submissions:
        if submission['workflow_task_id'] in fluidx_ids:
            ret.append({**submission,
                        'fluidx_id': fluidx_ids[submission['workflow_task_id']]})
        else:
            get_prefect_logger().warning('Cannot find Sanger Sample ID for workflow task: '
                                         + submission['workflow_task_id'])
    get_prefect_logger().info('Total number of viable submissions: ' + str(len(ret)))
    return ret


def get_created_dates_for_container_list(container_ids):
    benchling = get_benchling_instance()
    ret = {}
    # We can only get 20 at once
    for containers_page in [container_ids[i:i + 20] for i in range(0, len(container_ids), 20)]:
        container_results_page = generate_containers(
            benchling,
            ids=containers_page
        )
        for container in container_results_page:
            container_barcode = container.barcode
            created_date = container.created_at
            ret[container_barcode] = created_date.strftime('%Y-%m-%d %H:%M:%S')
    get_prefect_logger().info('Found this many container created dates: ' + str(len(ret)))
    return ret


@task(max_retries=3, retry_delay=timedelta(seconds=60))
def add_container_dates(submissions):
    container_ids = [d['container_eln_id'] for d in submissions]
    container_dates = get_created_dates_for_container_list(
        container_ids)
    ret = []
    for submission in submissions:
        if submission['sanger_sample_id'] in container_dates:
            ret.append({**submission,
                        'submission_date': container_dates[submission['sanger_sample_id']]})
        else:
            get_prefect_logger().warning('Cannot find created date for tube: '
                                         + submission['sanger_sample_id'])
    get_prefect_logger().info('Total number of viable submissions: ' + str(len(ret)))
    return ret


def get_contents_for_container_list(container_ids):
    benchling = get_benchling_instance()
    ret = {}
    # We can only get 20 at once
    for containers_page in [container_ids[i:i + 20] for i in range(0, len(container_ids), 20)]:
        returned_page = generate_containers(
            benchling,
            ids=containers_page,
            archive_reason='ANY_ARCHIVED_OR_NOT_ARCHIVED'
        )
        for container in returned_page:
            if len(container.contents) > 0:
                container_fluidx_id = container.barcode
                entity_id = container.contents[0].entity.to_dict()['fields']['Tissue']['value']
                ret[container_fluidx_id] = entity_id
    get_prefect_logger().info('Found this many Entity IDs: ' + str(len(ret)))
    return ret


@task(max_retries=3, retry_delay=timedelta(seconds=60))
def add_container_contents(submissions):
    container_ids = [d['container_eln_id'] for d in submissions]
    entity_ids = get_contents_for_container_list(
        container_ids)
    ret = []
    for submission in submissions:
        if submission['fluidx_id'] in entity_ids:
            ret.append({**submission,
                        'entity_id': entity_ids[submission['fluidx_id']]})
        else:
            get_prefect_logger().warning('Cannot find Entity ID for tube: '
                                         + submission['fluidx_id'])
    get_prefect_logger().info('Total number of containers with contents: ' + str(len(ret)))
    return ret


def get_sts_ids_for_entity_list(entity_ids):
    benchling = get_benchling_instance()
    ret = {}
    # We can only get 20 at once
    for entities_page in [entity_ids[i:i + 20] for i in range(0, len(entity_ids), 20)]:
        returned_page = generate_custom_entities(
            benchling,
            ids=entities_page,
            archive_reason='ANY_ARCHIVED_OR_NOT_ARCHIVED'
        )
        for entity in returned_page:
            entity_id = entity.id
            sts_id = entity.fields['STS ID'].value
            ret[entity_id] = sts_id
    get_prefect_logger().info('Found this many STS IDs: ' + str(len(ret)))
    return ret


@task(max_retries=3, retry_delay=timedelta(seconds=60))
def add_entity_sts_ids(submissions):
    entity_ids = [d['entity_id'] for d in submissions]
    sts_ids = get_sts_ids_for_entity_list(
        entity_ids)
    ret = []
    for submission in submissions:
        if submission['entity_id'] in sts_ids:
            ret.append({**submission,
                        'sts_id': sts_ids[submission['entity_id']]})
        else:
            get_prefect_logger().warning('Cannot find STS ID for tube: '
                                         + submission['fluidx_id'])
    get_prefect_logger().info("Total number of containers with content parent's STS id: "
                              f'{len(ret)}')
    return ret


@task(max_retries=3, retry_delay=timedelta(seconds=60))
def post_ep_samples_to_sts(submissions):
    updated_count = 0
    for submission in submissions:
        submission_date = submission['submission_date']
        if submission_date is None:
            get_prefect_logger().warning(submission['fluidx_id']
                                         + ' does not have a submission date')
            submission_date = '1970-01-01 00:00:00'
        payload = {'fluidx_id': submission['fluidx_id'],
                   'sample_id': submission['sts_id'],
                   'type': 'DNA',
                   'extraction_date': submission_date}
        r = sts_requests.post(
            '/ep_samples/' + submission['fluidx_id'],
            json=payload
        )
        if r.ok:
            updated_count += 1
        else:
            get_prefect_logger().warning(
                f'A sample failed with code {r.status_code}, '
                f'and response {r.json()}, '
                f'containing data: {payload}'
            )
    get_prefect_logger().info('Total number of ep_samples posted: ' + str(updated_count))
    return True


@task(max_retries=3, retry_delay=timedelta(seconds=60))
def get_lastrun_datetime(key):
    lastrun_datetime = get_datetime_setting(key)
    if lastrun_datetime is None:
        get_prefect_logger().error(
            'Could not load the datetime of last run.'
        )
        raise FAIL()
    get_prefect_logger().info(f'Last run on {lastrun_datetime}')
    return lastrun_datetime


@task(max_retries=3, retry_delay=timedelta(seconds=60))
def update_lastrun_datetime(key, new_datetime, go):
    success = update_datetime_setting(
        key,
        new_datetime
    )
    if not success:
        get_prefect_logger().error(
            'Could not update the datetime for this run.'
        )
        raise FAIL()
    get_prefect_logger().info(f'Updated last run date to {new_datetime}')
    return True


@task(max_retries=3, retry_delay=timedelta(seconds=60))
def get_new_lres_sample_data_from_eln(lastrun_datetime, schema_id):
    benchling = get_benchling_instance()
    lres_submissions = generate_workflow_outputs(
        benchling,
        schema_id=schema_id,
        modified_at='> ' + lastrun_datetime.isoformat('T')
    )
    lres_list = []
    for lres_submission in lres_submissions:
        tube_field = lres_submission.fields.to_dict()['Sample Tube ID']
        submission_date_field = lres_submission.fields.to_dict()['Submitted (Submission date)']
        submission_date = submission_date_field['value']
        if submission_date is not None:
            submission_date += ' 00:00:00'
        lres_list.append({'fluidx_id': sanitise_value(tube_field['displayValue']),
                          'container_eln_id': tube_field['value'],
                          'submission_date': submission_date})
    get_prefect_logger().info('Found this many LRES submissions: ' + str(len(lres_list)))
    return lres_list



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from collections.abc import Mapping
from typing import Callable, Iterator, Optional

from .converter import (
    S3Converter
)
from .parser import DefaultParser
from .s3_datasource import (
    S3ConverterFactory,
    S3DataSource
)
from ..core import DataSource
from ..services.s3_client import S3Client


class _S3DSDict(Mapping):
    def __init__(self, s3_ds: S3DataSource) -> None:
        self.__ds = s3_ds

    def __getitem__(self, __k: str) -> S3DataSource:
        if __k not in self.__ds.supported_types:
            raise KeyError()
        return self.__ds

    def __iter__(self) -> Iterator[str]:
        return iter(self.__ds.supported_types)

    def __len__(self) -> int:
        return len(self.__ds.supported_types)


class _ConverterFactory:
    """
    Manages the instantation of:

    - `S3Converter`
    """

    def __init__(self) -> None:
        self.__data_source: Optional[DataSource] = None

    @property
    def data_source(self) -> Optional[DataSource]:
        return self.__data_source

    @data_source.setter
    def data_source(
        self,
        ds: DataSource
    ) -> None:

        self.__data_source = ds

    def s3_converter_factory(self) -> S3ConverterFactory:
        """
        Returns an instantiated `S3Converter`.
        """

        parser = DefaultParser(self.__ds_dict)
        return S3Converter(parser)

    @property
    def __ds_dict(self) -> dict[str, DataSource]:
        return _S3DSDict(self.data_source)


def _get_client_factory() -> Callable[[], S3Client]:
    """
    A resonable default for creating
    an `S3Client` instance
    """

    return lambda: S3Client()


def create_s3_datasource(
    bucket_name: str,
    prefix: str = None
) -> S3DataSource:
    """
    Instantiates `S3DataSource` using the given:

    - `bucket_name`
    """

    client_factory = _get_client_factory()
    manager = _ConverterFactory()
    s3_ds = S3DataSource(
        client_factory,
        manager.s3_converter_factory,
        bucket_name,
        prefix
    )

    manager.data_source = s3_ds

    return s3_ds
